"""Tests for dataset.stats()."""

from datetime import datetime
from typing import Any, cast

import pytest
from pytest_mock import MockerFixture

from ..schema import Field, Item, MapType, field, schema
from . import dataset as dataset_module
from .dataset import StatsResult
from .dataset_test_utils import TestDataMaker

SIMPLE_ITEMS: list[Item] = [
  {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 1.0},
  {'float': float('nan')},
]


def test_simple_stats(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  result = dataset.stats(leaf_path='str')
  assert result == StatsResult(
    path=('str',), total_count=3, approx_count_distinct=2, avg_text_length=1
  )

  result = dataset.stats(leaf_path='float')
  assert result == StatsResult(
    path=('float',), total_count=4, approx_count_distinct=4, min_val=1.0, max_val=3.0
  )

  result = dataset.stats(leaf_path='bool')
  assert result == StatsResult(path=('bool',), total_count=3, approx_count_distinct=2)

  result = dataset.stats(leaf_path='int')
  assert result == StatsResult(
    path=('int',), total_count=3, approx_count_distinct=2, min_val=1, max_val=2
  )


def test_stats_with_deletions(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  result = dataset.stats(leaf_path='float')
  assert result == StatsResult(
    path=('float',), total_count=4, approx_count_distinct=4, min_val=1.0, max_val=3.0
  )

  dataset.delete_rows(filters=[('float', 'equals', 3.0)])

  result = dataset.stats(leaf_path='float')
  assert result == StatsResult(
    path=('float',), total_count=3, approx_count_distinct=3, min_val=1.0, max_val=2.0
  )

  dataset.restore_rows(filters=[('float', 'equals', 3.0)])

  result = dataset.stats(leaf_path='float')
  assert result == StatsResult(
    path=('float',), total_count=4, approx_count_distinct=4, min_val=1.0, max_val=3.0
  )


def test_nested_stats(make_test_data: TestDataMaker) -> None:
  nested_items: list[Item] = [
    {'name': 'Name1', 'addresses': [{'zips': [5, 8]}]},
    {'name': 'Name2', 'addresses': [{'zips': [3]}, {'zips': [11, 8]}]},
    {'name': 'Name2', 'addresses': []},  # No addresses.
    {'name': 'Name2', 'addresses': [{'zips': []}]},  # No zips in the first address.
  ]
  nested_schema = schema({'name': 'string', 'addresses': [{'zips': ['int32']}]})
  dataset = make_test_data(nested_items, schema=nested_schema)

  result = dataset.stats(leaf_path='name')
  assert result == StatsResult(
    path=('name',), total_count=4, approx_count_distinct=2, avg_text_length=5
  )

  result = dataset.stats(leaf_path='addresses.*.zips.*')
  assert result == StatsResult(
    path=('addresses', '*', 'zips', '*'),
    total_count=5,
    approx_count_distinct=4,
    min_val=3,
    max_val=11,
  )


def test_stats_approximation(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  sample_size = 5
  mocker.patch(f'{dataset_module.__name__}.TOO_MANY_DISTINCT', sample_size)

  nested_items: list[Item] = [{'feature': str(i)} for i in range(sample_size * 10)]
  nested_schema = schema({'feature': 'string'})
  dataset = make_test_data(nested_items, schema=nested_schema)

  result = dataset.stats(leaf_path='feature')
  assert result == StatsResult(
    path=('feature',), total_count=50, approx_count_distinct=50, avg_text_length=1
  )


def test_error_handling(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  with pytest.raises(ValueError, match='leaf_path must be provided'):
    dataset.stats(cast(Any, None))

  with pytest.raises(ValueError, match="Path \\('unknown',\\) not found in schema"):
    dataset.stats(leaf_path='unknown')


def test_map_dtype(make_test_data: TestDataMaker) -> None:
  items = [
    {'column': {'a': 1.0, 'b': 2.0}},
    {'column': {'b': 2.5}},
    {'column': {'a': 3.0, 'c': 3.5}},
  ]
  data_schema = schema(
    {'column': Field(dtype=MapType(key_type='string', value_field=field('float32')))}
  )
  dataset = make_test_data(items, schema=data_schema)
  assert dataset.stats('column.a') == StatsResult(
    path=('column', 'a'), total_count=2, approx_count_distinct=2, min_val=1.0, max_val=3.0
  )
  assert dataset.stats('column.b') == StatsResult(
    path=('column', 'b'), total_count=2, approx_count_distinct=2, min_val=2.0, max_val=2.5
  )
  assert dataset.stats('column.c') == StatsResult(
    path=('column', 'c'), total_count=1, approx_count_distinct=1, min_val=3.5, max_val=3.5
  )
  with pytest.raises(ValueError, match='Cannot compute stats on a map field'):
    dataset.stats('column')


def test_datetime(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'id': '1', 'date': datetime(2023, 1, 1)},
    {'id': '2', 'date': datetime(2023, 1, 15)},
    {'id': '2', 'date': datetime(2023, 2, 1)},
    {'id': '4', 'date': datetime(2023, 3, 1)},
    {
      'id': '5'
      # Missing datetime.
    },
  ]
  dataset = make_test_data(items)
  result = dataset.stats('date')
  assert result == StatsResult(
    path=('date',),
    total_count=4,
    approx_count_distinct=4,
    min_val=datetime(2023, 1, 1),
    max_val=datetime(2023, 3, 1),
  )



"""SQLite source."""

import os
import sqlite3
from typing import ClassVar, Iterable, Optional, cast

import duckdb
import pyarrow as pa
from fastapi import APIRouter
from pydantic import Field
from typing_extensions import override

from ..schema import Item, arrow_schema_to_schema
from ..source import Source, SourceSchema
from ..utils import file_exists
from .duckdb_utils import convert_path_to_duckdb, duckdb_setup

router = APIRouter()


@router.get('/tables')
def get_tables(db_file: str) -> list[str]:
  """List the table names in sqlite."""
  if not file_exists(db_file) or os.path.isdir(db_file):
    return []
  conn = sqlite3.connect(db_file)
  cursor = conn.cursor()
  cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
  res = [row[0] for row in cursor.fetchall()]
  cursor.close()
  conn.close()
  return res


class SQLiteSource(Source):
  """SQLite data loader."""

  name: ClassVar[str] = 'sqlite'
  router: ClassVar[APIRouter] = router

  db_file: str = Field(title='Database file', description='Path to the database file.')
  table: str = Field(description='Table name to read from.')

  _source_schema: Optional[SourceSchema] = None
  _reader: Optional[pa.RecordBatchReader] = None
  _con: Optional[duckdb.DuckDBPyConnection] = None

  @override
  def setup(self) -> None:
    self._con = duckdb.connect(database=':memory:')
    duckdb_setup(self._con)

    # DuckDB expects s3 protocol: https://duckdb.org/docs/guides/import/s3_import.html.
    duckdb_path = convert_path_to_duckdb(self.db_file)
    self._con.execute(
      f"""
      CREATE VIEW t as (SELECT * FROM sqlite_scan('{duckdb_path}', '{self.table}'));
    """
    )

    res = self._con.execute('SELECT COUNT(*) FROM t').fetchone()
    num_items = cast(tuple[int], res)[0]

    self._reader = self._con.execute('SELECT * from t').fetch_record_batch(rows_per_batch=10_000)
    # Create the source schema in prepare to share it between process and source_schema.
    schema = arrow_schema_to_schema(self._reader.schema)
    self._source_schema = SourceSchema(fields=schema.fields, num_items=num_items)

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    assert self._source_schema is not None
    return self._source_schema

  @override
  def yield_items(self) -> Iterable[Item]:
    """Process the source."""
    if not self._reader or not self._con:
      raise RuntimeError('SQLite source is not initialized.')

    for batch in self._reader:
      yield from batch.to_pylist()

    self._reader.close()
    self._con.close()



"""Item: an individual entry in the dataset."""

import csv
import functools
import io
from collections import deque
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Literal, Optional, Protocol, Sequence, Union, cast

import numpy as np
import pyarrow as pa
from pydantic import (
  BaseModel,
  ConfigDict,
  StrictInt,
  StrictStr,
  ValidationInfo,
  field_validator,
  model_validator,
)
from pydantic.functional_validators import ModelWrapValidatorHandler
from typing_extensions import TypedDict

from .utils import log

MANIFEST_FILENAME = 'manifest.json'
PARQUET_FILENAME_PREFIX = 'data'

# We choose `__rowid__` inspired by the standard `rowid` pseudocolumn in DBs:
# https://docs.oracle.com/cd/B19306_01/server.102/b14200/pseudocolumns008.htm
ROWID = '__rowid__'
PATH_WILDCARD = '*'
SPAN_KEY = '__span__'
VALUE_KEY = '__value__'
SIGNAL_METADATA_KEY = '__metadata__'
TEXT_SPAN_START_FEATURE = 'start'
TEXT_SPAN_END_FEATURE = 'end'

EMBEDDING_KEY = 'embedding'

# Python doesn't work with recursive types. These types provide some notion of type-safety.
Scalar = Union[bool, datetime, int, float, str, bytes]
Item = Any

# Contains a string field name, a wildcard for repeateds, or a specific integer index for repeateds.
# This path represents a path to a particular column.
# Examples:
#  ['article', 'field'] represents {'article': {'field': VALUES}}
#  ['article', '*', 'field'] represents {'article': [{'field': VALUES}, {'field': VALUES}]}
#  ['article', '0', 'field'] represents {'article': {'field': VALUES}}
PathTuple = tuple[str, ...]
Path = Union[PathTuple, str]

PathKeyedItem = tuple[Path, Item]

# These fields are for for python only and not written to a schema.
RichData = Union[str, bytes]
VectorKey = tuple[Union[StrictStr, StrictInt], ...]
PathKey = VectorKey


_SUPPORTED_PRIMITIVE_DTYPES = [
  'string',
  'string_span',
  'boolean',
  'int8',
  'int16',
  'int32',
  'int64',
  'uint8',
  'uint16',
  'uint32',
  'uint64',
  'float16',
  'float32',
  'float64',
  'time',
  'date',
  'timestamp',
  'interval',
  'binary',
  'embedding',
  'null',
  'map',
]


class DataType(BaseModel):
  """The data type for a field."""

  type: Literal[
    'string',
    'string_span',
    'boolean',
    'int8',
    'int16',
    'int32',
    'int64',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
    'float16',
    'float32',
    'float64',
    'time',
    'date',
    'timestamp',
    'interval',
    'binary',
    'embedding',
    'null',
    'map',
  ]

  def __init__(self, type: str, **kwargs: Any) -> None:
    super().__init__(type=type, **kwargs)

  @field_validator('type')
  @classmethod
  def _type_must_be_supported(cls, type: str) -> str:
    assert type in _SUPPORTED_PRIMITIVE_DTYPES, f'Unsupported type: {type}'
    return type

  @model_validator(mode='wrap')  # type: ignore
  def convert_str_to_dtype(v: Any, handler: ModelWrapValidatorHandler['DataType']) -> 'DataType':
    """Convert a string to a DataType."""
    if isinstance(v, str):
      v = {'type': v}
    return handler(v)

  def __str__(self) -> str:
    return self.type

  def __repr__(self) -> str:
    return self.type


def change_const_to_enum(prop_name: str, value: str) -> Callable[[dict[str, Any]], None]:
  """Replace the const value in the schema to an enum with 1 value so typescript codegen works."""

  def _schema_extra(schema: dict[str, Any]) -> None:
    schema['properties'][prop_name] = {'enum': [value]}
    if 'required' not in schema:
      schema['required'] = []
    schema['required'].append(prop_name)

  return _schema_extra


STRING = DataType('string')
# Contains {start, end} offset integers with a reference_column.
STRING_SPAN = DataType('string_span')
BOOLEAN = DataType('boolean')

# Ints.
INT8 = DataType('int8')
INT16 = DataType('int16')
INT32 = DataType('int32')
INT64 = DataType('int64')
UINT8 = DataType('uint8')
UINT16 = DataType('uint16')
UINT32 = DataType('uint32')
UINT64 = DataType('uint64')

# Floats.
FLOAT16 = DataType('float16')
FLOAT32 = DataType('float32')
FLOAT64 = DataType('float64')

### Time ###
# Time of day (no time zone).
TIME = DataType('time')
# Calendar date (year, month, day), no time zone.
DATE = DataType('date')
# An "Instant" stored as number of microseconds (µs) since 1970-01-01 00:00:00+00 (UTC time zone).
TIMESTAMP = DataType('timestamp')
# Time span, stored as microseconds.
INTERVAL = DataType('interval')

BINARY = DataType('binary')

EMBEDDING = DataType('embedding')

NULL = DataType('null')


class SignalInputType(str, Enum):
  """Enum holding the signal input type."""

  TEXT = 'text'
  TEXT_EMBEDDING = 'text_embedding'
  IMAGE = 'image'
  ANY = 'any'

  def __repr__(self) -> str:
    return self.value


EmbeddingInputType = Literal['question', 'document']


SIGNAL_TYPE_TO_VALID_DTYPES: dict[SignalInputType, list[DataType]] = {
  SignalInputType.TEXT: [STRING, STRING_SPAN],
  SignalInputType.IMAGE: [BINARY],
}


def signal_type_supports_dtype(input_type: SignalInputType, dtype: DataType) -> bool:
  """Returns True if the signal compute type supports the dtype."""
  if input_type == SignalInputType.ANY:
    return True
  return dtype in SIGNAL_TYPE_TO_VALID_DTYPES[input_type]


Bin = tuple[str, Optional[Union[float, int]], Optional[Union[float, int]]]


class MapFn(Protocol):
  """A map function.

  The map function will get called with the following arguments:
    - item: The item to be processed.
    - job_id: The job id.
  """

  __name__: str

  __call__: Callable[..., Item]


class MapInfo(BaseModel):
  """Holds information about a map that was run on a dataset."""

  fn_name: str
  input_path: Optional[PathTuple] = None
  fn_source: str
  date_created: datetime


class ClusterInputFormatSelectorInfo(BaseModel):
  """Information for a format selector for a cluster call."""

  format: str
  selector: str


class ClusterInfo(BaseModel):
  """Holds information about clustering operation that was run on a dataset."""

  min_cluster_size: Optional[int] = None
  remote: Optional[bool] = None
  input_path: Optional[PathTuple] = None
  input_format_selector: Optional[ClusterInputFormatSelectorInfo] = None


class MapType(DataType):
  """The map dtype parameterized by the key and value types."""

  def __init__(self, **kwargs: Any) -> None:
    kwargs.pop('type', None)
    super().__init__(type='map', **kwargs)

  type: Literal['map'] = 'map'
  key_type: DataType
  value_field: 'Field'

  model_config = ConfigDict(json_schema_extra=change_const_to_enum('type', 'map'))

  def __str__(self) -> str:
    return f'map<{self.key_type}, {self.value_field}>'


class Field(BaseModel):
  """Holds information for a field in the schema."""

  repeated_field: Optional['Field'] = None
  fields: Optional[dict[str, 'Field']] = None
  dtype: Optional[Union[MapType, DataType]] = None
  # Defined as the serialized signal when this field is the root result of a signal.
  signal: Optional[dict[str, Any]] = None
  # Defined as the label name when the field is a label.
  label: Optional[str] = None
  # Defined when the field is generated by a map.
  map: Optional[MapInfo] = None
  # Maps a named bin to a tuple of (start, end) values.
  bins: Optional[list[Bin]] = None
  categorical: Optional[bool] = None
  cluster: Optional[ClusterInfo] = None

  @field_validator('fields')
  @classmethod
  def validate_fields(cls, fields: Optional[dict[str, 'Field']]) -> Optional[dict[str, 'Field']]:
    """Validate the fields."""
    if not fields:
      return fields
    if VALUE_KEY in fields:
      raise ValueError(f'{VALUE_KEY} is a reserved field name.')
    return fields

  @field_validator('bins')
  @classmethod
  def validate_bins(cls, bins: list[Bin]) -> list[Bin]:
    """Validate the bins."""
    if len(bins) < 2:
      raise ValueError('Please specify at least two bins.')
    _, first_start, _ = bins[0]
    if first_start is not None:
      raise ValueError('The first bin should have a `None` start value.')
    _, _, last_end = bins[-1]
    if last_end is not None:
      raise ValueError('The last bin should have a `None` end value.')
    for i, (_, start, _) in enumerate(bins):
      if i == 0:
        continue
      prev_bin = bins[i - 1]
      _, _, prev_end = prev_bin
      if start != prev_end:
        raise ValueError(
          f'Bin {i} start ({start}) should be equal to the previous bin end {prev_end}.'
        )
    return bins

  @field_validator('categorical')
  @classmethod
  def validate_categorical(cls, categorical: bool, info: ValidationInfo) -> bool:
    """Validate the categorical field."""
    if categorical and is_float(info.data['dtype']):
      raise ValueError('Categorical fields cannot be float dtypes.')
    return categorical

  @model_validator(mode='after')
  def validate_field(self) -> 'Field':
    """Validate the field model."""
    if self.fields and self.repeated_field:
      raise ValueError('Both "fields" and "repeated_field" should not be defined')

    if self.dtype and self.repeated_field:
      raise ValueError('dtype and repeated_field cannot both be defined.')

    if not self.repeated_field and not self.fields and not self.dtype:
      raise ValueError('One of "fields", "repeated_field", or "dtype" should be defined')

    return self

  def __str__(self) -> str:
    return _str_field(self, indent=0)

  def __repr__(self) -> str:
    return f' {self.__class__.__name__}::{self.model_dump_json(exclude_none=True, indent=2)}'


class Schema(BaseModel):
  """Database schema."""

  # Map of top-level column names to fields (which are potentially complex structs).
  # Use .all_fields to get all fields, not just the top-level fields.
  fields: dict[str, Field]
  # Cached leafs.
  _leafs: Optional[dict[PathTuple, Field]] = None
  # Cached flat list of all the fields.
  _all_fields: Optional[list[tuple[PathTuple, Field]]] = None
  model_config = ConfigDict(arbitrary_types_allowed=True)

  @functools.cached_property
  def leafs(self) -> dict[PathTuple, Field]:
    """Return all the leaf fields in the schema. A leaf is defined as a node that contains a value.

    NOTE: Leafs may contain children. Leafs can be found as any node that has a dtype defined.
    """
    return dict((path, field) for path, field in self.all_fields if field.dtype)

  @functools.cached_property
  def all_fields(self) -> list[tuple[PathTuple, Field]]:
    """Return all the fields, including nested and repeated fields, as a flat list."""
    result: list[tuple[PathTuple, Field]] = []
    q: deque[tuple[PathTuple, Field]] = deque([((), Field(fields=self.fields))])
    while q:
      path, field = q.popleft()
      if path:
        result.append((path, field))
      if field.fields:
        for name, child_field in field.fields.items():
          child_path = (*path, name)
          q.append((child_path, child_field))
      elif field.repeated_field:
        child_path = (*path, PATH_WILDCARD)
        q.append((child_path, field.repeated_field))

    return result

  def __eq__(self, other: Any) -> bool:
    if not isinstance(other, Schema):
      return False
    return self.model_dump() == other.model_dump()

  def has_field(self, path: PathTuple) -> bool:
    """Returns if the field is found at the given path."""
    try:
      self.get_field(path)
      return True
    except ValueError:
      return False

  def get_field(self, path: PathTuple) -> Field:
    """Returns the field at the given path."""
    if path == (ROWID,):
      return Field(dtype=STRING)
    f = cast(Field, self)
    for name in path:
      if f.fields:
        if name not in f.fields:
          raise ValueError(f'Path {path} not found in schema')
        f = f.fields[name]
      elif f.repeated_field:
        if name != PATH_WILDCARD:
          raise ValueError(f'Invalid path for a schema field: {path}')
        f = f.repeated_field
      elif f.dtype and isinstance(f.dtype, MapType) and f.dtype.key_type == STRING:
        f = f.dtype.value_field
      else:
        raise ValueError(f'Invalid path for a schema field: {path}')
    return f

  def __str__(self) -> str:
    return _str_fields(self.fields, indent=0)

  def __repr__(self) -> str:
    return self.model_dump_json(exclude_none=True, indent=2)


def schema(schema_like: object) -> Schema:
  """Parse a schema-like object to a Schema object."""
  field = _parse_field_like(schema_like)
  if not field.fields:
    raise ValueError('Schema must have fields')
  return Schema(fields=field.fields)


def field(
  dtype: Optional[Union[DataType, str]] = None,
  signal: Optional[dict] = None,
  fields: Optional[object] = None,
  bins: Optional[list[Bin]] = None,
  categorical: Optional[bool] = None,
  label: Optional[str] = None,
  map: Optional[MapInfo] = None,
  cluster: Optional[ClusterInfo] = None,
) -> Field:
  """Parse a field-like object to a Field object."""
  field = _parse_field_like(fields or {}, dtype)
  if signal:
    field.signal = signal
  if map:
    field.map = map
  if cluster:
    field.cluster = cluster
  if label:
    field.label = label
  if dtype:
    if isinstance(dtype, str):
      dtype = DataType(dtype)
    field.dtype = dtype
  if bins:
    field.bins = bins
  if categorical is not None:
    field.categorical = categorical
  return field


class SpanVector(TypedDict):
  """A span with a vector."""

  span: tuple[int, int]
  vector: np.ndarray


def span(start: int, end: int, metadata: dict[str, Any] = {}) -> Item:
  """Creates a lilac span item, representing a pointer to a slice of text."""
  return {SPAN_KEY: {TEXT_SPAN_START_FEATURE: start, TEXT_SPAN_END_FEATURE: end}, **metadata}


def lilac_embedding(start: int, end: int, embedding: Optional[np.ndarray]) -> Item:
  """Creates a lilac embedding item, representing a vector with a pointer to a slice of text."""
  # Cast to int; we've had issues where start/end were np.int64, which caused downstream sadness.
  return span(int(start), int(end), {EMBEDDING_KEY: embedding})


def _parse_field_like(field_like: object, dtype: Optional[Union[DataType, str]] = None) -> Field:
  if isinstance(field_like, Field):
    return field_like
  elif isinstance(field_like, dict):
    fields: dict[str, Field] = {}
    for k, v in field_like.items():
      fields[k] = _parse_field_like(v)
    if isinstance(dtype, str):
      dtype = DataType(dtype)
    return Field(fields=fields or None, dtype=dtype)
  elif isinstance(field_like, str):
    return Field(dtype=DataType(field_like))
  elif isinstance(field_like, list):
    return Field(repeated_field=_parse_field_like(field_like[0], dtype=dtype))
  else:
    raise ValueError(f'Cannot parse field like: {field_like}')


def column_paths_match(path_match: Path, specific_path: Path) -> bool:
  """Test whether two column paths match.

  Args:
    path_match: A column path that contains wildcards, and sub-paths. This path will be used for
       testing the second specific path.
    specific_path: A column path that specifically identifies an field.

  Returns:
    Whether specific_path matches the path_match. This will only match when the
    paths are equal length. If a user wants to enrich everything with an array, they must use the
    path wildcard '*' in their patch match.
  """
  if isinstance(path_match, str):
    path_match = (path_match,)
  if isinstance(specific_path, str):
    specific_path = (specific_path,)

  if len(path_match) != len(specific_path):
    return False

  for path_match_p, specific_path_p in zip(path_match, specific_path):
    if path_match_p == PATH_WILDCARD:
      continue

    if path_match_p != specific_path_p:
      return False

  return True


class ImageInfo(BaseModel):
  """Info about an individual image."""

  path: Path


def normalize_path(path: Path) -> PathTuple:
  """Normalizes a dot seperated path, but ignores dots inside quotes, like regular SQL.

  Examples:
  - 'a.b.c' will be parsed as ('a', 'b', 'c').
  - '"a.b".c' will be parsed as ('a.b', 'c').
  - '"a".b.c' will be parsed as ('a', 'b', 'c').
  """
  if isinstance(path, str):
    return tuple(next(csv.reader(io.StringIO(path), delimiter='.')))
  return path


def _str_fields(fields: dict[str, Field], indent: int) -> str:
  prefix = ' ' * indent
  out: list[str] = []
  for name, field in fields.items():
    out.append(f'{prefix}{name}:{_str_field(field, indent=indent + 2)}')
  return '\n'.join(out)


def _str_field(field: Field, indent: int) -> str:
  if field.fields:
    prefix = '\n' if indent > 0 else ''
    return f'{prefix}{_str_fields(field.fields, indent)}'
  if field.repeated_field:
    return f' list({_str_field(field.repeated_field, indent)})'
  return f' {field.dtype}'


def dtype_to_arrow_schema(dtype: Optional[DataType]) -> Union[pa.Schema, pa.DataType]:
  """Convert the dtype to an arrow dtype."""
  if dtype == STRING:
    return pa.string()
  elif dtype == BOOLEAN:
    return pa.bool_()
  elif dtype == FLOAT16:
    return pa.float16()
  elif dtype == FLOAT32:
    return pa.float32()
  elif dtype == FLOAT64:
    return pa.float64()
  elif dtype == INT8:
    return pa.int8()
  elif dtype == INT16:
    return pa.int16()
  elif dtype == INT32:
    return pa.int32()
  elif dtype == INT64:
    return pa.int64()
  elif dtype == UINT8:
    return pa.uint8()
  elif dtype == UINT16:
    return pa.uint16()
  elif dtype == UINT32:
    return pa.uint32()
  elif dtype == UINT64:
    return pa.uint64()
  elif dtype == BINARY:
    return pa.binary()
  elif dtype == TIME:
    return pa.time64()
  elif dtype == DATE:
    return pa.date64()
  elif dtype == TIMESTAMP:
    return pa.timestamp('us')
  elif dtype == INTERVAL:
    return pa.duration('us')
  elif dtype == EMBEDDING:
    # We reserve an empty column for embeddings in parquet files so they can be queried.
    # The values are *not* filled out. If parquet and duckdb support embeddings in the future, we
    # can set this dtype to the relevant pyarrow type.
    return pa.null()
  elif dtype == STRING_SPAN:
    return pa.struct(
      {
        SPAN_KEY: pa.struct(
          {TEXT_SPAN_START_FEATURE: pa.int32(), TEXT_SPAN_END_FEATURE: pa.int32()}
        )
      }
    )
  elif dtype == NULL:
    return pa.null()
  elif dtype is None:
    return pa.null()
  elif dtype.type == 'map':
    map_dtype = cast(MapType, dtype)
    return pa.map_(
      dtype_to_arrow_schema(map_dtype.key_type), _schema_to_arrow_schema_impl(map_dtype.value_field)
    )
  else:
    raise ValueError(f'Can not convert dtype "{dtype}" to arrow dtype')


def schema_to_arrow_schema(schema: Union[Schema, Field]) -> pa.Schema:
  """Convert our schema to arrow schema."""
  arrow_schema = cast(pa.Schema, _schema_to_arrow_schema_impl(schema))
  arrow_fields = {field.name: field.type for field in arrow_schema}
  return pa.schema(arrow_fields)


def _schema_to_arrow_schema_impl(schema: Union[Schema, Field]) -> Union[pa.Schema, pa.DataType]:
  """Convert a schema to an apache arrow schema."""
  if schema.fields:
    arrow_fields: dict[str, Union[pa.Schema, pa.DataType]] = {}
    for name, field in schema.fields.items():
      if name == ROWID:
        arrow_schema = dtype_to_arrow_schema(field.dtype)
      else:
        arrow_schema = _schema_to_arrow_schema_impl(field)
      arrow_fields[name] = arrow_schema

    if isinstance(schema, Schema):
      return pa.schema(arrow_fields)
    else:
      # When nodes have both dtype and children, we add __value__ alongside the fields.
      if schema.dtype:
        value_schema = dtype_to_arrow_schema(schema.dtype)
        if schema.dtype == STRING_SPAN:
          arrow_fields[SPAN_KEY] = value_schema[SPAN_KEY].type
        else:
          arrow_fields[VALUE_KEY] = value_schema

      return pa.struct(arrow_fields)

  field = cast(Field, schema)
  if field.repeated_field:
    return pa.list_(_schema_to_arrow_schema_impl(field.repeated_field))

  return dtype_to_arrow_schema(field.dtype)


def arrow_dtype_to_dtype(arrow_dtype: pa.DataType) -> DataType:
  """Convert arrow dtype to our dtype."""
  # Ints.
  if arrow_dtype == pa.int8():
    return INT8
  elif arrow_dtype == pa.int16():
    return INT16
  elif arrow_dtype == pa.int32():
    return INT32
  elif arrow_dtype == pa.int64():
    return INT64
  elif arrow_dtype == pa.uint8():
    return UINT8
  elif arrow_dtype == pa.uint16():
    return UINT16
  elif arrow_dtype == pa.uint32():
    return UINT32
  elif arrow_dtype == pa.uint64():
    return UINT64
  # Floats.
  elif arrow_dtype == pa.float16():
    return FLOAT16
  elif arrow_dtype == pa.float32():
    return FLOAT32
  elif arrow_dtype == pa.float64():
    return FLOAT64
  # Time.
  elif pa.types.is_time(arrow_dtype):
    return TIME
  elif pa.types.is_date(arrow_dtype):
    return DATE
  elif pa.types.is_timestamp(arrow_dtype):
    return TIMESTAMP
  elif pa.types.is_duration(arrow_dtype):
    return INTERVAL
  # Others.
  elif arrow_dtype == pa.string():
    return STRING
  elif pa.types.is_binary(arrow_dtype) or pa.types.is_fixed_size_binary(arrow_dtype):
    return BINARY
  elif pa.types.is_boolean(arrow_dtype):
    return BOOLEAN
  elif arrow_dtype == pa.null():
    return NULL
  elif pa.types.is_map(arrow_dtype):
    return MapType(
      type='map',
      key_type=arrow_dtype_to_dtype(arrow_dtype.key_type),
      value_field=_arrow_schema_to_schema_impl(arrow_dtype.item_type),
    )
  else:
    raise ValueError(f'Can not convert arrow dtype "{arrow_dtype}" to our dtype')


def arrow_schema_to_schema(schema: pa.Schema) -> Schema:
  """Convert arrow schema to our schema."""
  # TODO(nsthorat): Change this implementation to allow more complicated reading of arrow schemas
  # into our schema by inferring values when {__value__: value} is present in the pyarrow schema.
  # This isn't necessary today as this util is only needed by sources which do not have data in the
  # lilac format.
  return cast(Schema, _arrow_schema_to_schema_impl(schema))


def _arrow_schema_to_schema_impl(schema: Union[pa.Schema, pa.DataType]) -> Union[Schema, Field]:
  """Convert an apache arrow schema to our schema."""
  if isinstance(schema, (pa.Schema, pa.StructType)):
    dtype: Optional[DataType] = None

    # Check for the span key, and start and end keys.
    if schema.get_field_index(SPAN_KEY) != -1 and not isinstance(schema, pa.Schema):
      span_schema = schema[SPAN_KEY].type
      if (
        isinstance(span_schema, (pa.Schema, pa.StructType))
        and span_schema.get_field_index(TEXT_SPAN_START_FEATURE) != -1
        and span_schema.get_field_index(TEXT_SPAN_END_FEATURE) != -1
        and pa.types.is_integer(span_schema[TEXT_SPAN_START_FEATURE].type)
        and pa.types.is_integer(span_schema[TEXT_SPAN_END_FEATURE].type)
      ):
        dtype = STRING_SPAN

    fields: dict[str, Field] = {
      field.name: cast(Field, _arrow_schema_to_schema_impl(field.type))
      for field in schema
      if field.name != SPAN_KEY
    }

    return (
      Schema(fields=fields) if isinstance(schema, pa.Schema) else Field(dtype=dtype, fields=fields)
    )
  elif isinstance(schema, pa.ListType):
    return Field(repeated_field=cast(Field, _arrow_schema_to_schema_impl(schema.value_field.type)))
  else:
    return Field(dtype=arrow_dtype_to_dtype(schema))


def is_float(dtype: DataType) -> bool:
  """Check if a dtype is a float dtype."""
  return dtype in [FLOAT16, FLOAT32, FLOAT64]


def is_integer(dtype: DataType) -> bool:
  """Check if a dtype is an integer dtype."""
  return dtype in [
    INT8,
    INT16,
    INT32,
    INT64,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
  ]


def is_temporal(dtype: DataType) -> bool:
  """Check if a dtype is a temporal dtype."""
  return dtype in [TIME, DATE, TIMESTAMP, INTERVAL]


def is_ordinal(dtype: DataType) -> bool:
  """Check if a dtype is an ordinal dtype."""
  return is_float(dtype) or is_integer(dtype) or is_temporal(dtype)


def _print_fields(field: Field, destination: Field) -> None:
  log('New field:')
  log(field)
  log('Destination field:')
  log(destination)


def _merge_field_into(field: Field, destination: Field, disallow_pedals: bool = False) -> None:
  destination.signal = destination.signal or field.signal
  destination.map = destination.map or field.map
  if field.fields:
    if destination.repeated_field:
      _print_fields(field, destination)
      raise ValueError(
        'Failed to merge fields. New field has fields, but destination is ' 'repeated'
      )
    if disallow_pedals and destination.dtype:
      _print_fields(field, destination)
      raise ValueError('Failed to merge fields. New field has fields, but destination has dtype')
    destination.fields = destination.fields or {}
    if destination.dtype == NULL:
      destination.dtype = None
    for field_name, subfield in field.fields.items():
      if field_name not in destination.fields:
        destination.fields[field_name] = subfield.model_copy(deep=True)
      else:
        _merge_field_into(subfield, destination.fields[field_name], disallow_pedals)
  elif field.repeated_field:
    if destination.fields:
      _print_fields(field, destination)
      raise ValueError(
        'Failed to merge fields. New field is repeated, but destination has ' 'fields'
      )
    if destination.dtype:
      if destination.dtype != NULL:
        _print_fields(field, destination)
        raise ValueError(
          'Failed to merge fields. New field is repeated, but destination has ' 'dtype'
        )
      else:
        destination.dtype = None
        destination.repeated_field = field.repeated_field.model_copy(deep=True)
        return
    if not destination.repeated_field:
      _print_fields(field, destination)
      raise ValueError('Failed to merge fields. New field is repeated, but destination is not')
    _merge_field_into(field.repeated_field, destination.repeated_field, disallow_pedals)

  if field.dtype:
    if field.dtype == NULL:
      return
    if destination.repeated_field:
      _print_fields(field, destination)
      raise ValueError(
        'Failed to merge fields. New field has dtype, but destination is ' 'repeated'
      )
    if disallow_pedals and destination.fields:
      _print_fields(field, destination)
      raise ValueError('Failed to merge fields. New field has dtype, but destination has fields')

    if (
      destination.dtype != NULL
      and destination.dtype is not None
      and destination.dtype != field.dtype
    ):
      raise ValueError(
        f'Failed to merge fields. New field has dtype {field.dtype}, but '
        f'destination has dtype {destination.dtype}'
      )
    destination.dtype = field.dtype


def merge_fields(fields: Sequence[Field], disallow_pedals: bool = False) -> Field:
  """Merge a list of fields.

  Args:
    fields: A list of fields to merge.
    disallow_pedals: If True, merging a field with dtype into a field with fields will raise an
                     error.
  """
  if not fields:
    return Field(dtype=NULL)
  merged_field = fields[0].model_copy(deep=True)
  for field in fields[1:]:
    _merge_field_into(field, merged_field, disallow_pedals)
  return merged_field


def merge_schemas(schemas: Sequence[Union[Schema, Field]]) -> Schema:
  """Merge a list of schemas."""
  merged_field = merge_fields([Field(fields=s.fields) for s in schemas])
  assert merged_field.fields, 'Merged field should have fields'
  return Schema(fields=merged_field.fields)



"""Test the Spacy NER signal."""

from ..schema import field
from ..splitters.text_splitter_test_utils import text_to_expected_spans
from .ner import SpacyNER


def test_spacy_ner_fields() -> None:
  signal = SpacyNER()
  signal.setup()
  assert signal.fields() == field(fields=[field('string_span', fields={'label': 'string'})])


def test_ner() -> None:
  signal = SpacyNER()
  signal.setup()

  text = (
    'Net income was $9.4 million compared to the prior year of $2.7 million.'
    'Revenue exceeded twelve billion dollars, with a loss of $1b.'
  )
  emails = list(signal.compute([text]))

  expected_spans = text_to_expected_spans(
    text,
    [
      ('$9.4 million', {'label': 'MONEY'}),
      ('the prior year', {'label': 'DATE'}),
      ('$2.7 million', {'label': 'MONEY'}),
      ('twelve billion dollars', {'label': 'MONEY'}),
      ('1b', {'label': 'MONEY'}),
    ],
  )

  assert emails == [expected_spans]



"""Finds markdown code blocks.

NOTE: It would be great to use guesslang to detect the language automatically, however
there is a dependency conflict with typing extensions.
"""
import re
from typing import ClassVar, Iterable, Iterator, Optional, cast

from typing_extensions import override

from ..schema import Field, Item, RichData, field, span
from ..signal import TextSignal

MARKDOWN_RE = '```([^\n ]*?)\n(.*?)\n```'


class MarkdownCodeBlockSignal(TextSignal):
  """Finds markdown blocks in text. Emits the language of the block with the span."""

  name: ClassVar[str] = 'markdown_code_block'
  display_name: ClassVar[str] = 'Markdown Code Block Detection'

  @override
  def fields(self) -> Field:
    return field(
      fields=[
        field(
          dtype='string_span',
          fields={'language': 'string'},
        )
      ]
    )

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    markdown_re = re.compile(MARKDOWN_RE, re.MULTILINE | re.DOTALL)
    for doc in data:
      text = cast(str, doc)
      # Get the spans
      markdown_re_spans = markdown_re.finditer(text)
      languages = markdown_re.findall(text)

      spans: list[Item] = []
      for re_span, (language, _) in zip(markdown_re_spans, languages):
        spans.append(span(re_span.start(), re_span.end(), {'language': language}))

      yield spans



"""Signals enrich a document with additional metadata."""

from ..signal import Signal, SignalInputType, TextEmbeddingSignal, TextSignal, register_signal
from .concept_scorer import ConceptSignal
from .default_signals import register_default_signals
from .lang_detection import LangDetectionSignal
from .near_dup import NearDuplicateSignal
from .ner import SpacyNER
from .pii import PIISignal
from .text_statistics import TextStatisticsSignal

register_default_signals()

__all__ = [
  'Signal',
  'TextEmbeddingSignal',
  'TextStatisticsSignal',
  'TextSignal',
  'register_signal',
  'SignalInputType',
  'LangDetectionSignal',
  'NearDuplicateSignal',
  'SpacyNER',
  'PIISignal',
  'ConceptSignal',
]



"""Startup work before running the web server on HuggingFace."""

import os
import shutil
from typing import TypedDict

import yaml

from .concepts.db_concept import DiskConceptDB, get_concept_output_dir
from .db_manager import get_dataset, list_datasets
from .env import env, get_project_dir
from .project import PROJECT_CONFIG_FILENAME
from .utils import DebugTimer, get_datasets_dir, get_lilac_cache_dir, log


def delete_old_files() -> None:
  """Delete old files from the cache."""
  try:
    from huggingface_hub import scan_cache_dir
  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )

  # Scan cache
  try:
    scan = scan_cache_dir()
  except BaseException:
    # Cache was not found.
    return

  # Select revisions to delete
  to_delete = []
  for repo in scan.repos:
    latest_revision = max(repo.revisions, key=lambda x: x.last_modified)
    to_delete.extend(
      [revision.commit_hash for revision in repo.revisions if revision != latest_revision]
    )
  strategy = scan.delete_revisions(*to_delete)

  # Delete them
  log(f'Will delete {len(to_delete)} old revisions and save {strategy.expected_freed_size_str}')
  strategy.execute()


class HfSpaceConfig(TypedDict):
  """The huggingface space config, defined in README.md.

  See:
  https://huggingface.co/docs/hub/spaces-config-reference
  """

  title: str
  datasets: list[str]


def hf_docker_start() -> None:
  """Download dataset files from the HF space that was uploaded before building the image."""
  log('Setting up data dependencies for the server...')
  with DebugTimer('Huggingface docker start'):
    _hf_docker_start()


def _hf_docker_start() -> None:
  try:
    from huggingface_hub import snapshot_download
  except ImportError:
    raise ImportError(
      'Could not import the "huggingface_hub" python package. '
      'Please install it with `pip install "huggingface_hub".'
    )

  # SPACE_ID is the HuggingFace Space ID environment variable that is automatically set by HF.
  repo_id = env('SPACE_ID', None)
  if not repo_id:
    return

  delete_old_files()

  with open(os.path.abspath('README.md')) as f:
    # Strip the '---' for the huggingface readme config.
    readme_contents = f.read()
    dashes = '---'
    first_dashes = readme_contents.find(dashes)
    second_dashes = readme_contents.find(dashes, first_dashes + len(dashes))

    readme = readme_contents[first_dashes + len(dashes) : second_dashes]
    hf_config: HfSpaceConfig = yaml.safe_load(readme)

  # Download the huggingface space data. This includes code and datasets, so we move the datasets
  # alone to the data directory.

  datasets_dir = get_datasets_dir(get_project_dir())
  os.makedirs(datasets_dir, exist_ok=True)
  for lilac_hf_dataset in hf_config.get('datasets', []):
    print('Downloading dataset from HuggingFace: ', lilac_hf_dataset)
    snapshot_download(
      repo_id=lilac_hf_dataset,
      repo_type='dataset',
      token=env('HF_ACCESS_TOKEN'),
      local_dir=datasets_dir,
      ignore_patterns=['.gitattributes', 'README.md'],
    )

  snapshot_dir = snapshot_download(repo_id=repo_id, repo_type='space', token=env('HF_ACCESS_TOKEN'))

  spaces_data_dir = os.path.join(snapshot_dir, 'data')
  # Copy the config file.
  project_config_file = os.path.join(spaces_data_dir, PROJECT_CONFIG_FILENAME)
  if os.path.exists(project_config_file):
    shutil.copy(project_config_file, os.path.join(get_project_dir(), PROJECT_CONFIG_FILENAME))

  # Delete cache files from persistent storage.
  cache_dir = get_lilac_cache_dir(get_project_dir())
  if os.path.exists(cache_dir):
    shutil.rmtree(cache_dir)

  # Copy cache files from the space if they exist.
  spaces_cache_dir = get_lilac_cache_dir(spaces_data_dir)
  if os.path.exists(spaces_cache_dir):
    shutil.copytree(spaces_cache_dir, cache_dir)

  # Copy concepts.
  concepts = DiskConceptDB(spaces_data_dir).list()
  for concept in concepts:
    # Ignore lilac concepts, they're already part of the source code.
    if concept.namespace == 'lilac':
      continue
    spaces_concept_output_dir = get_concept_output_dir(
      spaces_data_dir, concept.namespace, concept.name
    )
    persistent_output_dir = get_concept_output_dir(
      get_project_dir(), concept.namespace, concept.name
    )
    shutil.rmtree(persistent_output_dir, ignore_errors=True)
    shutil.copytree(spaces_concept_output_dir, persistent_output_dir, dirs_exist_ok=True)
    shutil.rmtree(spaces_concept_output_dir, ignore_errors=True)

  # Compute dataset indices, if indices are enabled:
  if env('USE_TABLE_INDEX', default=False):
    for dataset_info in list_datasets():
      ds_namespace, ds_name = dataset_info.namespace, dataset_info.dataset_name
      ds = get_dataset(ds_namespace, ds_name)
      ds.manifest()



"""Router for the concept database."""

from typing import Annotated, Iterable, Optional, cast

import instructor
from fastapi import APIRouter, HTTPException
from fastapi.params import Depends
from instructor import OpenAISchema
from pydantic import BaseModel, Field

from .auth import UserInfo, get_session_user
from .concepts.concept import (
  DRAFT_MAIN,
  Concept,
  ConceptMetadata,
  ConceptMetrics,
  ConceptType,
  DraftId,
  draft_examples,
)
from .concepts.db_concept import DISK_CONCEPT_DB, DISK_CONCEPT_MODEL_DB, ConceptInfo, ConceptUpdate
from .env import env
from .router_utils import RouteErrorHandler, server_compute_concept
from .schema import RichData
from .signals.concept_scorer import ConceptSignal

router = APIRouter(route_class=RouteErrorHandler)


@router.get('/', response_model_exclude_none=True)
def get_concepts(
  user: Annotated[Optional[UserInfo], Depends(get_session_user)]
) -> list[ConceptInfo]:
  """List the concepts."""
  return DISK_CONCEPT_DB.list(user)


@router.get('/{namespace}/{concept_name}', response_model_exclude_none=True)
def get_concept(
  namespace: str,
  concept_name: str,
  draft: Optional[DraftId] = DRAFT_MAIN,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)] = None,
) -> Concept:
  """Get a concept from a database."""
  concept = DISK_CONCEPT_DB.get(namespace, concept_name, user)
  if not concept:
    raise HTTPException(
      status_code=404,
      detail=f'Concept "{namespace}/{concept_name}" was not found or user does not have access.',
    )

  # Only return the examples from the draft.
  concept.data = draft_examples(concept, draft or DRAFT_MAIN)

  return concept


class CreateConceptOptions(BaseModel):
  """Options for creating a concept."""

  # Namespace of the concept.
  namespace: str
  # Name of the concept.
  name: str
  # Input type (modality) of the concept.
  type: ConceptType
  metadata: Optional[ConceptMetadata] = None


@router.post('/create', response_model_exclude_none=True)
def create_concept(
  options: CreateConceptOptions, user: Annotated[Optional[UserInfo], Depends(get_session_user)]
) -> Concept:
  """Edit a concept in the database."""
  return DISK_CONCEPT_DB.create(
    options.namespace, options.name, options.type, metadata=options.metadata, user=user
  )


@router.post('/{namespace}/{concept_name}', response_model_exclude_none=True)
def edit_concept(
  namespace: str,
  concept_name: str,
  change: ConceptUpdate,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> Concept:
  """Edit a concept in the database."""
  return DISK_CONCEPT_DB.edit(namespace, concept_name, change, user)


@router.post('/{namespace}/{concept_name}/metadata', response_model_exclude_none=True)
def edit_concept_metadata(
  namespace: str,
  concept_name: str,
  concept_metadata: ConceptMetadata,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> None:
  """Edit the metadata of a concept."""
  DISK_CONCEPT_DB.update_metadata(
    namespace=namespace, name=concept_name, metadata=concept_metadata, user=user
  )


@router.delete('/{namespace}/{concept_name}')
def delete_concept(
  namespace: str, concept_name: str, user: Annotated[Optional[UserInfo], Depends(get_session_user)]
) -> None:
  """Deletes the concept from the database."""
  DISK_CONCEPT_DB.remove(namespace, concept_name, user)


class MergeConceptDraftOptions(BaseModel):
  """Merge a draft into main."""

  draft: DraftId


@router.post('/{namespace}/{concept_name}/merge_draft', response_model_exclude_none=True)
def merge_concept_draft(
  namespace: str,
  concept_name: str,
  options: MergeConceptDraftOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> Concept:
  """Merge a draft in the concept into main."""
  return DISK_CONCEPT_DB.merge_draft(namespace, concept_name, options.draft, user)


class ScoreExample(BaseModel):
  """Example to score along a specific concept."""

  text: Optional[str] = None
  img: Optional[bytes] = None


class ScoreBody(BaseModel):
  """Request body for the score endpoint."""

  examples: list[ScoreExample]
  draft: str = DRAFT_MAIN


class ConceptModelInfo(BaseModel):
  """Information about a concept model."""

  namespace: str
  concept_name: str
  embedding_name: str
  version: int
  metrics: Optional[ConceptMetrics] = None


@router.get('/{namespace}/{concept_name}/model')
def get_concept_models(
  namespace: str,
  concept_name: str,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)] = None,
) -> list[ConceptModelInfo]:
  """Get a concept model from a database."""
  concept = DISK_CONCEPT_DB.get(namespace, concept_name, user)
  if not concept:
    raise HTTPException(
      status_code=404, detail=f'Concept "{namespace}/{concept_name}" was not found'
    )
  models = DISK_CONCEPT_MODEL_DB.get_models(namespace, concept_name, user)

  for m in models:
    DISK_CONCEPT_MODEL_DB.sync(m.namespace, m.concept_name, m.embedding_name, user)

  return [
    ConceptModelInfo(
      namespace=m.namespace,
      concept_name=m.concept_name,
      embedding_name=m.embedding_name,
      version=m.version,
      metrics=m.get_metrics(),
    )
    for m in models
  ]


@router.get('/{namespace}/{concept_name}/model/{embedding_name}')
def get_concept_model(
  namespace: str,
  concept_name: str,
  embedding_name: str,
  create_if_not_exists: bool = False,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)] = None,
) -> Optional[ConceptModelInfo]:
  """Get a concept model from a database."""
  concept = DISK_CONCEPT_DB.get(namespace, concept_name, user)
  if not concept:
    raise HTTPException(
      status_code=404, detail=f'Concept "{namespace}/{concept_name}" was not found'
    )

  model = DISK_CONCEPT_MODEL_DB.get(namespace, concept_name, embedding_name, user)
  if not model and not create_if_not_exists:
    return None

  model = DISK_CONCEPT_MODEL_DB.sync(
    namespace, concept_name, embedding_name, user=user, create=create_if_not_exists
  )
  model_info = ConceptModelInfo(
    namespace=model.namespace,
    concept_name=model.concept_name,
    embedding_name=model.embedding_name,
    version=model.version,
    metrics=model.get_metrics(),
  )
  return model_info


@router.post(
  '/{namespace}/{concept_name}/model/{embedding_name}/score', response_model_exclude_none=True
)
def score(
  namespace: str,
  concept_name: str,
  embedding_name: str,
  body: ScoreBody,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> list[list[dict]]:
  """Score examples along the specified concept."""
  concept_scorer = ConceptSignal(
    namespace=namespace, concept_name=concept_name, embedding=embedding_name
  )
  concept_scorer.set_user(user)
  return cast(
    list[list[dict]],
    server_compute_concept(
      concept_scorer, cast(Iterable[RichData], [e.text for e in body.examples]), user
    ),
  )


class Examples(OpenAISchema):
  """Generated text examples."""

  examples: list[str] = Field(..., description='List of generated examples')


@router.get('/generate_examples')
def generate_examples(description: str) -> list[str]:
  """Generate positive examples for a given concept using an LLM model."""
  api_key = env('OPENAI_API_KEY')
  api_type = env('OPENAI_API_TYPE')
  api_version = env('OPENAI_API_VERSION')
  api_engine = env('OPENAI_API_ENGINE_CHAT')
  if not api_key:
    raise ValueError('`OPENAI_API_KEY` environment variable not set.')
  try:
    import openai

  except ImportError:
    raise ImportError(
      'Could not import the "openai" python package. '
      'Please install it with `pip install openai`.'
    )
  else:
    openai.api_key = api_key
    api_engine = api_engine

    if api_type:
      openai.api_type = api_type
      openai.api_version = api_version

  try:
    # Enables response_model in the openai client.
    client = instructor.patch(openai.OpenAI())

    completion = client.chat.completions.create(
      model='gpt-3.5-turbo',
      functions=[Examples.openai_schema],
      messages=[
        {
          'role': 'system',
          'content': 'You must call the `Examples` function with the generated examples',
        },
        {
          'role': 'user',
          'content': f'Write 5 diverse, unnumbered, and concise examples of "{description}"',
        },
      ],
    )
    result = Examples.from_response(completion)
    return result.examples

  except openai.AuthenticationError:
    raise ValueError(
      'Your `OPENAI_API_KEY` environment variable need to be completed with '
      '`OPENAI_API_TYPE`, `OPENAI_API_BASE`, `OPENAI_API_VERSION`, `OPENAI_API_ENGINE_CHAT`'
    )



"""Tests for the CSV source."""
import csv
import os
import pathlib

import pytest
from pytest_mock import MockerFixture

from ..schema import schema
from ..source import SourceSchema
from ..test_utils import retrieve_parquet_rows
from .csv_source import LINE_NUMBER_COLUMN, CSVSource


@pytest.fixture(autouse=True)
def set_project_path(tmp_path: pathlib.Path, mocker: MockerFixture) -> None:
  # We need to set the project path so extensions like https will be installed in temp dir.
  mocker.patch.dict(os.environ, {'LILAC_PROJECT_DIR': str(tmp_path)})


def test_csv(tmp_path: pathlib.Path) -> None:
  csv_rows = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}, {'x': 3, 'y': 'thirty'}]

  filename = 'test-dataset.csv'
  filepath = os.path.join(tmp_path, filename)
  with open(filepath, 'w') as f:
    writer = csv.DictWriter(f, fieldnames=list(csv_rows[0].keys()))
    writer.writeheader()
    writer.writerows(csv_rows)

  source = CSVSource(filepaths=[filepath])
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({LINE_NUMBER_COLUMN: 'int64', 'x': 'int64', 'y': 'string'}).fields, num_items=3
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [
    {LINE_NUMBER_COLUMN: 1, 'x': 1, 'y': 'ten'},
    {LINE_NUMBER_COLUMN: 2, 'x': 2, 'y': 'twenty'},
    {LINE_NUMBER_COLUMN: 3, 'x': 3, 'y': 'thirty'},
  ]


def test_csv_multifile(tmp_path: pathlib.Path) -> None:
  csv_rows = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}, {'x': 3, 'y': 'thirty'}]

  filepaths = []
  for i in range(3):
    filename = f'test-dataset-{i}.csv'
    filepath = os.path.join(tmp_path, filename)
    with open(filepath, 'w') as f:
      writer = csv.DictWriter(f, fieldnames=list(csv_rows[0].keys()))
      writer.writeheader()
      writer.writerows(csv_rows)
    filepaths.append(filepath)

  source = CSVSource(filepaths=filepaths)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({LINE_NUMBER_COLUMN: 'int64', 'x': 'int64', 'y': 'string'}).fields, num_items=9
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert len(items) == 9



"""Embedding registry."""
from typing import Callable, Iterable, Optional, Union

import numpy as np
from pydantic import StrictStr

from ..schema import (
  EMBEDDING_KEY,
  SPAN_KEY,
  TEXT_SPAN_END_FEATURE,
  TEXT_SPAN_START_FEATURE,
  EmbeddingInputType,
  Item,
  RichData,
  SpanVector,
  lilac_embedding,
)
from ..signal import TextEmbeddingSignal, get_signal_by_type
from ..splitters.chunk_splitter import TextChunk
from ..utils import chunks

EMBEDDING_SORT_PRIORITIES = [
  'gte-small',
  'gte-base',
  'gte-large',
  'jina-v2-small',
  'jina-v2-base',
  'openai',
  'sbert',
]

EmbeddingId = Union[StrictStr, TextEmbeddingSignal]

EmbedFn = Callable[[Iterable[RichData]], Iterable[list[SpanVector]]]


def get_embed_fn(
  embedding_name: str, split: bool, input_type: EmbeddingInputType = 'document'
) -> EmbedFn:
  """Return a function that returns the embedding matrix for the given embedding signal."""
  embedding_cls = get_signal_by_type(embedding_name, TextEmbeddingSignal)
  embedding = embedding_cls(split=split, embed_input_type=input_type)
  embedding.setup()

  def _embed_fn(data: Iterable[RichData]) -> Iterable[list[SpanVector]]:
    items = embedding.compute(data)

    for item in items:
      if not item:
        raise ValueError('Embedding signal returned None.')

      yield [
        {
          'vector': item_val[EMBEDDING_KEY].reshape(-1),
          'span': (
            item_val[SPAN_KEY][TEXT_SPAN_START_FEATURE],
            item_val[SPAN_KEY][TEXT_SPAN_END_FEATURE],
          ),
        }
        for item_val in item
      ]

  return _embed_fn


def identity_chunker(doc: str) -> list[TextChunk]:
  """Split a document into a single chunk."""
  return [(doc, (0, len(doc)))]


def chunked_compute_embedding(
  embed_fn: Callable[[list[str]], list[np.ndarray]],
  docs: list[str],
  batch_size: int,
  chunker: Callable[[str], list[TextChunk]] = identity_chunker,
) -> list[Optional[list[Item]]]:
  """Compute text embeddings for chunks of text, using the provided splitter and embedding fn."""
  text_chunks = [(i, chunk) for i, doc in enumerate(docs) for chunk in chunker(doc)]
  output: list[list[Item]] = [[] for _ in docs]
  batches = chunks(text_chunks, batch_size)
  for batch in batches:
    batch_texts = [text for _, (text, _) in batch]
    batch_embeddings = embed_fn(batch_texts)
    for (i, (_, (start, end))), embedding in zip(batch, batch_embeddings):
      output[i].append(lilac_embedding(start, end, embedding))

  return [lis or None for lis in output]



"""SpaCy-based chunk splitting algorithms."""
import bisect
import functools
from typing import TYPE_CHECKING, Optional, Sequence

import numpy as np
import tiktoken

from .chunk_splitter import TextChunk

if TYPE_CHECKING:
  import spacy

# Vector size to use for chunk comparisons.
BOW_VECTOR_SIZE = 128


@functools.cache
def get_tokenizer() -> tiktoken.Encoding:
  """Lazily instantiate and return a singleton TikToken tokenizer object."""
  return tiktoken.get_encoding('cl100k_base')


def embed_tokenizer_BoW(texts: list[str]) -> np.ndarray:
  """Encode a list of texts as a bag-of-words vector."""
  tokenizer = get_tokenizer()
  out = np.zeros((len(texts), BOW_VECTOR_SIZE), dtype=np.float32)
  for i, text in enumerate(texts):
    token_ids = np.array(tokenizer.encode(text, disallowed_special=())) % BOW_VECTOR_SIZE
    out[i] = np.bincount(token_ids, minlength=BOW_VECTOR_SIZE)
  return out


@functools.cache
def get_spacy() -> 'spacy.Language':
  """Lazily instantiate and return a singeton SpaCy sentencizer object."""
  # Lazy import; spacy is a heavy dependency.
  try:
    import spacy
  except ImportError:
    raise ImportError(
      'Could not import the "spacy" python package. ' 'Please install it with `pip install spacy`.'
    )

  sentencizer = spacy.blank('en')
  # This includes colon as a sentence boundary; LLM datasets tend to contain a lot of semantic
  # markers with a colon, like "Teacher: ... " or "Answer the following question: ..."
  sentencizer.add_pipe('sentencizer', config={'punct_chars': [':', ';', '.', '!', '?']})
  # Increase the number of characters of the tokenizer as we're not using a parser or NER.
  sentencizer.max_length = 10_000_000
  return sentencizer


def simple_spacy_chunker(text: str, filter_short: int = 4) -> list[TextChunk]:
  """Split text into sentence-based chunks, using SpaCy."""
  sentencizer = get_spacy()
  chunks = [
    (text[s.start_char : s.end_char], (s.start_char, s.end_char)) for s in sentencizer(text).sents
  ]
  # Filter out stray whitespace, list numberings, etc.
  chunks = [c for c in chunks if len(c[0].strip()) > filter_short]
  return chunks


def group_by_embedding(
  fulltext: str, chunks: list[TextChunk], target_num_groups: int, max_len: int
) -> list[TextChunk]:
  """Take a series of smaller chunks and cluster them together.

  Args:
    fulltext: Full text.
    chunks: Smaller chunks to combine.
    embed_fn: A function mapping strings to an embedding vector.
    target_num_groups: Target number of chunks in final output.
    max_len: Maximum size of a combined chunk.
  """
  if not chunks:
    return []

  texts = [c[0] for c in chunks]
  embeddings = embed_tokenizer_BoW(texts)
  # Center the embeddings for all sentences; this accentuates sentence semantics,
  # especially if the entire passage is roughly about the same topic
  embeddings -= np.mean(embeddings, axis=0)
  # Add a small amount of noise to prevent division by zero.
  embeddings += np.random.uniform(size=embeddings.shape) * 1e-6
  embeddings /= np.linalg.norm(embeddings, axis=-1, keepdims=True)

  neighbor_distances: Sequence[float] = (embeddings[:-1] * embeddings[1:]).sum(axis=1)
  potential_breaks: np.ndarray = np.array([c[1][1] for c in chunks[:-1]])  # end index of each chunk
  priority_sort_breaks: np.ndarray = potential_breaks[np.argsort(neighbor_distances)]

  # If there are fewer sentences than target number of groups, then this should degrade gracefully.
  breakpoints = [0] + sorted(priority_sort_breaks[: (target_num_groups - 1)]) + [chunks[-1][1][1]]

  def _find_long_spans(breakpoints: list[int]) -> Optional[tuple[int, int]]:
    for i, j in zip(breakpoints[:-1], breakpoints[1:]):
      if j - i > max_len:
        return (i, j)
    return None

  # Recursively break long spans until there are no more.
  while (span := _find_long_spans(breakpoints)) is not None:
    i, j = span
    for potential_break in priority_sort_breaks:
      if i < potential_break < j:
        bisect.insort(breakpoints, potential_break)
        break
    else:  # No potential breaker was found. Arbitrarily split the span in half.
      bisect.insort(breakpoints, int((i + j) / 2))

  return [
    (fulltext[start:end], (start, end)) for start, end in zip(breakpoints[:-1], breakpoints[1:])
  ]


def clustering_spacy_chunker(
  text: str, filter_short: int = 4, max_len: int = 512, target_num_groups: Optional[int] = None
) -> list[TextChunk]:
  """Split text into sentence-based chunks, with semantic clustering to join related sentences."""
  chunks = simple_spacy_chunker(text, filter_short=filter_short)

  if target_num_groups is None:
    # A rough heuristic for picking a number of target chunks.
    # These magic numbers were chosen by manually chunking 40 texts spanning 50-5000 characters in
    # length, and eyeballing a best-fit line from #num chunks vs. #length on a log-log plot.
    target_num_groups = max(1, int((len(text) ** 0.33) / 1.5))
  return group_by_embedding(text, chunks, target_num_groups, max_len)



"""Utils for the python server."""
import gzip
import itertools
import json
from io import BytesIO
from typing import Any, Callable, Generator, Iterable, Iterator, TypeVar, Union, cast

from .schema import PATH_WILDCARD, Item, PathTuple
from .utils import chunks, is_primitive


def _flatten_iter(input: Union[Iterator, object], max_depth: int) -> Iterator:
  """Flattens a nested iterable."""
  if max_depth == 0 or isinstance(input, dict) or is_primitive(input):
    yield input
  else:
    for elem in cast(Iterator, input):
      yield from _flatten_iter(elem, max_depth - 1)


def flatten_iter(input: Union[Iterator, Iterable], max_depth: int = -1) -> Iterator:
  """Flattens a deeply nested iterator. Primitives and dictionaries are not flattened."""
  for elem in input:
    yield from _flatten_iter(elem, max_depth)


def flatten_path_iter(input: Union[Iterator, Iterable, dict], path: PathTuple) -> Iterator:
  """Flattens a nested object along the provided path."""
  if not path:
    yield input
    return
  path_part = path[0]
  rest = path[1:]
  if path_part == PATH_WILDCARD:
    for elem in cast(Iterator, input):
      yield from flatten_path_iter(elem, rest)
  elif path_part not in input:
    return
  else:
    yield from flatten_path_iter(cast(dict, input)[path_part], rest)


def _unflatten_iter(
  flat_input: Iterator[list[object]], original_input: Union[Iterable, object], max_depth: int
) -> Union[list, dict]:
  """Unflattens a deeply flattened iterable according to the original iterable's structure."""
  if max_depth == 0 or is_primitive(original_input) or isinstance(original_input, dict):
    return next(flat_input)
  else:
    values = cast(Iterable, original_input)
    return [_unflatten_iter(flat_input, orig_elem, max_depth - 1) for orig_elem in values]


def unflatten_iter(
  flat_input: Union[Iterable, Iterator],
  original_input: Union[Iterable, object],
  max_depth: int = -1,
) -> Generator:
  """Unflattens a deeply flattened iterable according to the original iterable's structure."""
  flat_input_iter = iter(flat_input)
  if isinstance(original_input, Iterable) and not is_primitive(original_input):
    for o in original_input:
      yield _unflatten_iter(flat_input_iter, o, max_depth)
    return

  yield _unflatten_iter(iter(flat_input), original_input, max_depth)


TFlatten = TypeVar('TFlatten')


def flatten(inputs: Iterable[Iterable[TFlatten]]) -> Iterator[TFlatten]:
  """Flattens a nested iterator.

  Only supports flattening one level deep.
  """
  for input in inputs:
    yield from input


TUnflatten = TypeVar('TUnflatten')


def unflatten(
  flat_inputs: Union[Iterable[TUnflatten], Iterator[TUnflatten]],
  original_inputs: Iterable[Iterable[Any]],
) -> Iterator[list[TUnflatten]]:
  """Unflattens a flattened iterable according to the original iterable's structure."""
  flat_inputs_iter = iter(flat_inputs)
  for original_input in original_inputs:
    yield [next(flat_inputs_iter) for _ in original_input]


TFlatBatchedInput = TypeVar('TFlatBatchedInput')
TFlatBatchedOutput = TypeVar('TFlatBatchedOutput')


def group_by_sorted_key_iter(
  input: Iterator[Item], key_fn: Callable[[Item], object]
) -> Iterator[list[Item]]:
  """Takes a key-sorted iterator and yields each group of items sharing the same key.

  Args:
    input: An iterator of items, ordered by the key.
    key_fn: A function that takes an item and returns a key.

  Yields:
    A list of items sharing the same key, for each group.
  """
  last_key: object = None
  last_group: list[Item] = []
  for item in input:
    key = key_fn(item)
    if key != last_key:
      if last_group:
        yield last_group
      last_group = [item]
      last_key = key
    else:
      last_group.append(item)
  if last_group:
    yield last_group


def flat_batched_compute(
  input: Iterable[Iterable[TFlatBatchedInput]],
  f: Callable[[list[TFlatBatchedInput]], Iterable[TFlatBatchedOutput]],
  batch_size: int,
) -> Iterator[Iterable[TFlatBatchedOutput]]:
  """Flatten the input, batched call f, and return the output unflattened."""
  # Tee the input so we can use it twice for the input and output shapes.
  input_1, input_2 = itertools.tee(input, 2)
  batches = chunks(flatten(input_1), batch_size)
  batched_outputs = flatten((f(batch) for batch in batches))
  return unflatten(batched_outputs, input_2)


def compress_docs(docs: list[str]) -> bytes:
  """Compress a list of documents using gzip."""
  buffer = BytesIO()
  # Gzip compress the data and write it to the buffer
  with gzip.GzipFile(fileobj=buffer, mode='wb', compresslevel=6) as f:
    f.write(json.dumps(docs).encode('utf-8'))
  return buffer.getvalue()



"""Gegeral Text Embeddings (GTE) model. Open-source model, designed to run on device."""
import gc
from typing import TYPE_CHECKING, ClassVar, Optional

from typing_extensions import override

if TYPE_CHECKING:
  from sentence_transformers import SentenceTransformer

from ..schema import Item
from ..signal import TextEmbeddingSignal
from ..splitters.spacy_splitter import clustering_spacy_chunker
from ..tasks import TaskExecutionType
from .embedding import chunked_compute_embedding
from .transformer_utils import SENTENCE_TRANSFORMER_BATCH_SIZE, setup_model_device

# See https://huggingface.co/spaces/mteb/leaderboard for leaderboard of models.
GTE_SMALL = 'thenlper/gte-small'
GTE_BASE = 'thenlper/gte-base'
GTE_TINY = 'TaylorAI/gte-tiny'


class GTESmall(TextEmbeddingSignal):
  """Computes Gegeral Text Embeddings (GTE).

  <br>This embedding runs on-device. See the [model card](https://huggingface.co/thenlper/gte-small)
  for details.
  """

  name: ClassVar[str] = 'gte-small'
  display_name: ClassVar[str] = 'Gegeral Text Embeddings (small)'
  local_batch_size: ClassVar[int] = SENTENCE_TRANSFORMER_BATCH_SIZE
  local_parallelism: ClassVar[int] = 1
  local_strategy: ClassVar[TaskExecutionType] = 'threads'

  _model_name = GTE_SMALL
  _model: 'SentenceTransformer'

  @override
  def setup(self) -> None:
    try:
      from sentence_transformers import SentenceTransformer
    except ImportError:
      raise ImportError(
        'Could not import the "sentence_transformers" python package. '
        'Please install it with `pip install "sentence_transformers".'
      )
    self._model = setup_model_device(SentenceTransformer(self._model_name), self._model_name)

  @override
  def compute(self, docs: list[str]) -> list[Optional[Item]]:
    """Call the embedding function."""
    # While we get docs in batches of 1024, the chunker expands that by a factor of 3-10.
    # The sentence transformer API actually does batching internally, so we pass
    # local_batch_size * 16 to allow the library to see all the chunks at once.
    return chunked_compute_embedding(
      self._model.encode, docs, self.local_batch_size * 16, chunker=clustering_spacy_chunker
    )

  @override
  def teardown(self) -> None:
    self._model.cpu()
    del self._model
    gc.collect()

    try:
      import torch

      torch.cuda.empty_cache()
    except ImportError:
      pass


class GTEBase(GTESmall):
  """Computes Gegeral Text Embeddings (GTE).

  <br>This embedding runs on-device. See the [model card](https://huggingface.co/thenlper/gte-base)
  for details.
  """

  name: ClassVar[str] = 'gte-base'
  display_name: ClassVar[str] = 'Gegeral Text Embeddings (base)'

  _model_name = GTE_BASE


class GTETiny(GTESmall):
  """Computes Gegeral Text Embeddings (GTE)."""

  name: ClassVar[str] = 'gte-tiny'
  display_name: ClassVar[str] = 'Gegeral Text Embeddings (tiny)'

  _model_name = GTE_TINY



"""Tests utils of for dataset_test."""
import json
import os
import pathlib
from copy import deepcopy
from datetime import datetime
from typing import Any, ClassVar, Optional, Type, cast

import numpy as np
from typing_extensions import Protocol

from ..config import DatasetConfig
from ..embeddings.vector_store import VectorDBIndex
from ..project import add_project_dataset_config, create_project, update_project_dataset_settings
from ..schema import (
  BINARY,
  BOOLEAN,
  FLOAT32,
  INT32,
  MANIFEST_FILENAME,
  NULL,
  PARQUET_FILENAME_PREFIX,
  ROWID,
  SPAN_KEY,
  STRING,
  STRING_SPAN,
  TIMESTAMP,
  VALUE_KEY,
  DataType,
  Field,
  Item,
  PathKey,
  Schema,
  merge_fields,
)
from ..source import Source, SourceManifest
from ..utils import get_dataset_output_dir, is_primitive, open_file
from .dataset import Dataset, default_settings
from .dataset_utils import write_items_to_parquet

TEST_NAMESPACE = 'test_namespace'
TEST_DATASET_NAME = 'test_dataset'


class TestDataMaker(Protocol):
  """A function that creates a test dataset."""

  def __call__(self, items: list[Item], schema: Optional[Schema] = None) -> Dataset:
    """Create a test dataset."""
    ...


class TestSource(Source):
  """Test source that does nothing."""

  name: ClassVar[str] = 'test_source'


def _infer_dtype(value: Item) -> DataType:
  if isinstance(value, str):
    return STRING
  elif isinstance(value, bool):
    return BOOLEAN
  elif isinstance(value, bytes):
    return BINARY
  elif isinstance(value, float):
    return FLOAT32
  elif isinstance(value, int):
    return INT32
  elif isinstance(value, datetime):
    return TIMESTAMP
  else:
    raise ValueError(f'Cannot infer dtype of primitive value: {value}')


def _infer_field(item: Item, diallow_pedals: bool = False) -> Field:
  """Infer the schema from the items."""
  if isinstance(item, dict):
    fields: dict[str, Field] = {}
    for k, v in item.items():
      fields[k] = _infer_field(cast(Item, v))
    dtype = None
    if VALUE_KEY in fields:
      dtype = fields[VALUE_KEY].dtype
      del fields[VALUE_KEY]
    elif SPAN_KEY in fields:
      dtype = STRING_SPAN
      del fields[SPAN_KEY]
    if not fields:
      # The object is an empty dict. We need a dummy child to represent this with parquet.
      return Field(dtype=NULL)
    return Field(fields=fields, dtype=dtype)
  elif is_primitive(item):
    return Field(dtype=_infer_dtype(item))
  elif isinstance(item, list):
    inferred_fields = [_infer_field(subitem) for subitem in item]
    merged_field = merge_fields(inferred_fields, diallow_pedals)
    return Field(repeated_field=merged_field)
  else:
    raise ValueError(f'Cannot infer schema of item: {item}')


def _infer_schema(items: list[Item]) -> Schema:
  """Infer the schema from a list of items."""
  merged_field = merge_fields(
    [_infer_field(item, diallow_pedals=True) for item in items], disallow_pedals=True
  )
  if not merged_field.fields:
    raise ValueError(f'Failed to infer schema. Got {merged_field}')
  return Schema(fields=merged_field.fields)


def make_dataset(
  dataset_cls: Type[Dataset],
  tmp_path: pathlib.Path,
  items: list[Item],
  schema: Optional[Schema] = None,
) -> Dataset:
  """Create a test dataset."""
  schema = schema or _infer_schema(items)
  _write_items(tmp_path, TEST_DATASET_NAME, items, schema)
  create_project(str(tmp_path))

  config = DatasetConfig(namespace=TEST_NAMESPACE, name=TEST_DATASET_NAME, source=TestSource())
  add_project_dataset_config(config)

  dataset = dataset_cls(TEST_NAMESPACE, TEST_DATASET_NAME)
  settings = default_settings(dataset)
  update_project_dataset_settings(TEST_NAMESPACE, TEST_DATASET_NAME, settings)

  return dataset


def _write_items(
  tmpdir: pathlib.Path, dataset_name: str, items: list[Item], schema: Schema
) -> None:
  """Write the items JSON to the dataset format: manifest.json and parquet files."""
  source_dir = get_dataset_output_dir(str(tmpdir), TEST_NAMESPACE, dataset_name)
  os.makedirs(source_dir)

  # Add rowids to the items.
  items = [deepcopy(item) for item in items]
  for i, item in enumerate(items):
    item[ROWID] = str(i + 1)

  simple_parquet_files = write_items_to_parquet(
    items, source_dir, schema, filename_prefix=PARQUET_FILENAME_PREFIX, shard_index=0, num_shards=1
  )
  manifest = SourceManifest(files=[simple_parquet_files], data_schema=schema, source=TestSource())
  with open_file(os.path.join(source_dir, MANIFEST_FILENAME), 'w') as f:
    f.write(manifest.model_dump_json(indent=2, exclude_none=True))


def enriched_item(value: Optional[Item] = None, metadata: dict[str, Item] = {}) -> Item:
  """Wrap a value in a dict with the value key."""
  return {VALUE_KEY: value, **metadata}


def make_vector_index(
  vector_store: str, vector_dict: dict[PathKey, list[list[float]]]
) -> VectorDBIndex:
  """Make a vector index from a dictionary of vector keys to vectors."""
  embeddings: list[np.ndarray] = []
  spans: list[tuple[PathKey, list[tuple[int, int]]]] = []
  for path_key, vectors in vector_dict.items():
    vector_spans: list[tuple[int, int]] = []
    for i, vector in enumerate(vectors):
      embeddings.append(np.array(vector))
      vector_spans.append((0, 0))
    spans.append((path_key, vector_spans))

  vector_index = VectorDBIndex(vector_store)
  vector_index.add(spans, np.array(embeddings))
  return vector_index


class TestProcessLogger:
  """A simple utility for unit tests that need logging between processes and main process."""

  def __init__(self, tmp_dir: pathlib.Path):
    self.tmp_dir = tmp_dir

  def log_event(self, value: Any) -> None:
    """Logs an event to the log file for communicate between worker & test."""
    with open(os.path.join(self.tmp_dir, 'log.jsonl'), 'a') as f:
      f.write(f'{json.dumps(value)}\n')

  def clear_logs(self) -> None:
    """Clears the logs."""
    with open(os.path.join(self.tmp_dir, 'log.jsonl'), 'w') as f:
      f.write('')

  def get_logs(self) -> list[Any]:
    """Gets the logs."""
    with open(os.path.join(self.tmp_dir, 'log.jsonl'), 'r') as f:
      return [json.loads(line) for line in f.readlines()]



"""Compute text statistics for a document."""
from typing import TYPE_CHECKING, ClassVar, Iterable, Optional, cast

from typing_extensions import override

from ..schema import Field, Item, field
from ..signal import TextSignal

SPACY_LANG_MODEL = 'en_core_web_sm'
SPACY_BATCH_SIZE = 128
SPACY_MAX_LENGTH = 2_000_000

NUM_CHARS = 'num_characters'
READABILITY = 'readability'
TYPE_TOKEN_RATIO = 'log(type_token_ratio)'
FRAC_NON_ASCII = 'frac_non_ascii'

if TYPE_CHECKING:
  from spacy import Language
  from spacy.tokens import Doc


class TextStatisticsSignal(TextSignal):
  """Compute text statistics for a document such as readability scores, type-token-ratio, etc.."""

  name: ClassVar[str] = 'text_statistics'
  display_name: ClassVar[str] = 'Text Statistics'
  local_batch_size: ClassVar[int] = SPACY_BATCH_SIZE

  _lang: Optional['Language'] = None

  @override
  def fields(self) -> Field:
    return field(
      fields={
        NUM_CHARS: 'int32',
        READABILITY: 'float32',
        TYPE_TOKEN_RATIO: 'float32',
        FRAC_NON_ASCII: field(
          'float32', bins=[('Low', None, 0.15), ('Medium', 0.15, 0.3), ('High', 0.3, None)]
        ),
      }
    )

  @override
  def setup(self) -> None:
    try:
      import spacy
      import spacy.cli
      import spacy.util
    except ImportError:
      raise ImportError(
        'Could not import the "spacy" python package. '
        'Please install it with `pip install spacy`.'
      )

    if not spacy.util.is_package(SPACY_LANG_MODEL):
      spacy.cli.download(SPACY_LANG_MODEL)
    self._lang = spacy.load(
      SPACY_LANG_MODEL,
      disable=[
        'parser',
        'tagger',
        'ner',
        'lemmatizer',
        'textcat',
        'custom',
        'tok2vec',
        'attribute_ruler',
      ],
    )
    self._lang.max_length = SPACY_MAX_LENGTH

  @override
  def compute(self, docs: list[str]) -> list[Optional[Item]]:
    try:
      import textacy.corpus
      from textacy import text_stats
    except ImportError:
      raise ImportError(
        'Could not import the "textacy" python package. '
        'Please install it with `pip install textacy`.'
      )
    if not self._lang:
      raise RuntimeError('Language model was not loaded.')

    # Replace None with empty strings to avoid spacy errors.
    docs = [x or '' for x in docs]
    # See https://textacy.readthedocs.io/en/0.11.0/api_reference/text_stats.html for a list of
    # available statistics.
    return_data: list[Optional[Item]] = []
    corpus = textacy.corpus.Corpus(lang=self._lang, data=docs)
    for doc in cast(Iterable['Doc'], corpus):
      if not doc or not doc.text.strip():
        return_data.append(None)
        continue
      try:
        readability = text_stats.readability.automated_readability_index(doc)
      except ZeroDivisionError:
        readability = None
      try:
        ttr = text_stats.diversity.log_ttr(doc)
      except ValueError:
        ttr = None
      num_chars = len(doc.text)
      num_non_ascii = 0
      for c in doc.text:
        if ord(c) >= 128:
          num_non_ascii += 1
      frac_non_ascii = num_non_ascii / num_chars if num_chars else 0

      return_data.append(
        {
          NUM_CHARS: num_chars,
          READABILITY: readability,
          TYPE_TOKEN_RATIO: ttr,
          FRAC_NON_ASCII: frac_non_ascii,
        }
      )
    return return_data



"""Concepts are customizable signals that help enrich documents."""

from .concept import Example, ExampleIn
from .db_concept import ConceptUpdate, DiskConceptDB, DiskConceptModelDB

__all__ = ['DiskConceptDB', 'DiskConceptModelDB', 'Example', 'ExampleIn', 'ConceptUpdate']



"""Load environment variables from .env file."""
import os
import pathlib
from typing import Any, Optional, Union

from dotenv import load_dotenv
from pydantic import BaseModel
from pydantic import Field as PydanticField

# The project directory defaults to the current working directory.
DEFAULT_PROJECT_DIR = '.'


# NOTE: This is created for documentation, but isn't parsed by pydantic until we update to 2.0.
class LilacEnvironment(BaseModel):
  """Lilac environment variables.

  These can be set with operating system environment variables to override behavior.

  For python, see: https://docs.python.org/3/library/os.html#os.environ

  For bash, see: https://www.gnu.org/software/bash/manual/bash.html#Environment
  """

  # General Lilac environment variables.
  LILAC_DATA_PATH: str = PydanticField(
    description='[Deprecated] The Lilac data path where datasets, concepts, caches are stored. '
    'This is deprecated in favor of `LILAC_PROJECT_DIR`, but will work for backwards compat.'
  )
  LILAC_PROJECT_DIR: str = PydanticField(
    description='The Lilac project directory where datasets, concepts, caches are stored.'
    'This replaces `LILAC_PROJECT_DIR`, which is deprecated but as the same functionality. '
    'This can be set with `set_project_dir`.'
  )

  DEBUG: str = PydanticField(
    description='Turn on Lilac debug mode to log queries and timing information.'
  )
  DISABLE_LOGS: str = PydanticField(description='Disable log() statements to the console.')
  USE_TABLE_INDEX: str = PydanticField(description='Use persistent tables with rowid indexes.')

  # API Keys.
  OPENAI_API_KEY: str = PydanticField(
    description='The OpenAI API key, used for computing `openai` embeddings and generating '
    'positive examples for concept seeding.'
  )
  COHERE_API_KEY: str = PydanticField(
    description='The Cohere API key, used for computing `cohere` embeddings.'
  )
  PALM_API_KEY: str = PydanticField(
    description='The PaLM API key, used for computing `palm` embeddings.'
  )

  # HuggingFace demo.
  HF_ACCESS_TOKEN: str = PydanticField(
    description='The HuggingFace access token, used for downloading data to a space from a '
    'private dataset. This is also required if the HuggingFace space is private.'
  )

  # Authentication.
  LILAC_AUTH_ENABLED: str = PydanticField(
    description='Set to true to enable read-only mode, disabling the ability to add datasets & '
    'compute dataset signals. When enabled, `GOOGLE_CLIENT_ID`, `GOOGLE_CLIENT_SECRET` and '
    '`LILAC_OAUTH_SECRET_KEY` should also be set.'
  )

  LILAC_AUTH_ADMIN_EMAILS: str = PydanticField(
    description='A comma-separated list of Google emails that are allowed full edit-access, as if '
    'the `LILAC_AUTH_ENABLED` environment flag was disabled. These email addresses are used in '
    'concert with the `GOOGLE_CLIENT_ID` and `GOOGLE_CLIENT_SECRET` environment flags to '
    'authenticate users.'
  )
  LILAC_AUTH_USER_EDIT_LABELS: str = PydanticField(
    description='Set to true to allow non-admin users to edit labels.'
  )
  LILAC_AUTH_USER_DISABLE_LABEL_ALL: str = PydanticField(
    description='Set to true to disable non-admin users to use the label-all feature in the UI.'
  )

  GOOGLE_CLIENT_ID: str = PydanticField(
    description='The Google OAuth client ID. Required when `LILAC_AUTH_ENABLED=true`. Details can '
    'be found at https://developers.google.com/identity/protocols/oauth2.'
  )
  GOOGLE_CLIENT_SECRET: str = PydanticField(
    description='The Google OAuth client secret. Details can be found at '
    'https://developers.google.com/identity/protocols/oauth2.'
  )
  LILAC_OAUTH_SECRET_KEY: str = PydanticField(
    description='The Google OAuth random secret key. Details can be found at '
    'https://developers.google.com/identity/protocols/oauth2.'
  )

  # Other settings.
  GOOGLE_ANALYTICS_ENABLED: str = PydanticField(
    description='Set to to true to enable Google analytics.'
  )
  LILAC_LOAD_ON_START_SERVER: str = PydanticField(
    description='When true, will load from lilac.yml upon startup.'
  )

  GCS_REGION: str = PydanticField(description='The GCS region for GCS operations.')
  GCS_ACCESS_KEY: str = PydanticField(description='The GCS access key for GCS operations.')
  GCS_SECRET_KEY: str = PydanticField(description='The GCS secret key for GCS operations.')

  S3_REGION: str = PydanticField(description='The S3 region for S3 operations.')
  S3_ACCESS_KEY: str = PydanticField(description='The S3 access key for S3 operations.')
  S3_SECRET_KEY: str = PydanticField(description='The S3 secret key for S3 operations.')
  S3_ENDPOINT: str = PydanticField(
    description='The S3 endpoint URL for S3-like operations, including GCS and Azure.'
  )


def _init_env() -> None:
  in_test = os.environ.get('LILAC_TEST', None)
  # Load the .env files into the environment in order of highest to lowest priority.

  if not in_test:  # Skip local environment variables when testing.
    load_dotenv('.env.local')
  load_dotenv('.env')

  auth_enabled = os.environ.get('LILAC_AUTH_ENABLED', False) == 'true'
  if auth_enabled:
    if not os.environ.get('GOOGLE_CLIENT_ID', None) or not os.environ.get(
      'GOOGLE_CLIENT_SECRET', None
    ):
      raise ValueError(
        'Missing `GOOGLE_CLIENT_ID` or `GOOGLE_CLIENT_SECRET` when `LILAC_AUTH_ENABLED=true`'
      )
    SECRET_KEY = os.environ.get('LILAC_OAUTH_SECRET_KEY', None)
    if not SECRET_KEY:
      raise ValueError('Missing `LILAC_OAUTH_SECRET_KEY` when `LILAC_AUTH_ENABLED=true`')
  if auth_enabled:
    if not os.environ.get('GOOGLE_CLIENT_ID', None) or not os.environ.get(
      'GOOGLE_CLIENT_SECRET', None
    ):
      raise ValueError(
        'Missing `GOOGLE_CLIENT_ID` or `GOOGLE_CLIENT_SECRET` when `LILAC_AUTH_ENABLED=true`'
      )
    SECRET_KEY = os.environ.get('LILAC_OAUTH_SECRET_KEY', None)
    if not SECRET_KEY:
      raise ValueError('Missing `LILAC_OAUTH_SECRET_KEY` when `LILAC_AUTH_ENABLED=true`')


def env(key: str, default: Optional[Any] = None) -> Any:
  """Return the value of an environment variable."""
  return os.environ.get(key, default)


def get_project_dir() -> str:
  """Return the base path for data."""
  project_dir = env('LILAC_PROJECT_DIR', None)
  if not project_dir:
    project_dir = env('LILAC_DATA_PATH', DEFAULT_PROJECT_DIR)
  if not project_dir:
    raise ValueError('`LILAC_PROJECT_DIR` environment variable must be set. ')
  return project_dir


def set_project_dir(project_dir: Union[str, pathlib.Path]) -> None:
  """Set the project directory."""
  project_dir = os.path.expanduser(project_dir)
  os.makedirs(project_dir, exist_ok=True)
  os.environ['LILAC_PROJECT_DIR'] = str(project_dir)


# Initialize the environment at import time.
_init_env()



"""Authentication and ACL configuration."""

from typing import Optional

import modal.config
from fastapi import Request
from pydantic import BaseModel, ValidationError

from .env import env


class ConceptAuthorizationException(Exception):
  """Authorization exceptions thrown by the concept database."""

  pass


class DatasetUserAccess(BaseModel):
  """User access for datasets."""

  # Whether the user can compute a signal.
  compute_signals: bool
  # Whether the user can delete a dataset.
  delete_dataset: bool
  # Whether the user can delete a signal.
  delete_signals: bool
  # Whether the user can update settings.
  update_settings: bool
  # Whether the user can create a new label type.
  create_label_type: bool
  # Whether the user can add or remove labels.
  edit_labels: bool
  # Whether the user can use the label all feature.
  label_all: bool
  # Whether the user can delete rows.
  delete_rows: bool
  # Whether the user can execute jobs remotely.
  execute_remotely: bool


class ConceptUserAccess(BaseModel):
  """User access for concepts."""

  # Whether the user can delete any concept (not their own).
  delete_any_concept: bool


class UserAccess(BaseModel):
  """User access."""

  is_admin: bool = False

  create_dataset: bool

  # TODO(nsthorat): Make this keyed to each dataset and concept.
  dataset: DatasetUserAccess
  concept: ConceptUserAccess


class UserInfo(BaseModel):
  """User information."""

  id: str
  email: str
  name: str
  given_name: str
  family_name: str


class AuthenticationInfo(BaseModel):
  """Authentication information for the user."""

  user: Optional[UserInfo] = None
  access: UserAccess
  auth_enabled: bool
  # The HuggingFace space ID if the server is running on a HF space.
  huggingface_space_id: Optional[str] = None


def has_garden_credentials() -> bool:
  """Returns whether the user has Garden credentials."""
  config = modal.config.Config().to_dict()
  return (
    'token_secret' in config
    and 'token_id' in config
    and 'lilacai' in modal.config.config_profiles()
  )


def get_session_user(request: Request) -> Optional[UserInfo]:
  """Get the user from the session."""
  if not env('LILAC_AUTH_ENABLED'):
    return None
  user_info_dict = request.session.get('user', None)
  if user_info_dict:
    try:
      return UserInfo.model_validate(user_info_dict)
    except ValidationError:
      return None
  return None


def get_admin_emails() -> list[str]:
  """Return the admin emails."""
  admin_emails = env('LILAC_AUTH_ADMIN_EMAILS', None)
  if admin_emails:
    return admin_emails.split(',')
  return []


def get_user_access(user_info: Optional[UserInfo]) -> UserAccess:
  """Get the user access."""
  auth_enabled = env('LILAC_AUTH_ENABLED')
  if isinstance(auth_enabled, str):
    auth_enabled = auth_enabled.lower() == 'true'

  admin_emails = get_admin_emails()
  is_admin = not auth_enabled or (user_info is not None and user_info.email in admin_emails)

  if auth_enabled and not is_admin:
    return UserAccess(
      is_admin=is_admin,
      create_dataset=False,
      dataset=DatasetUserAccess(
        compute_signals=False,
        delete_dataset=False,
        delete_signals=False,
        update_settings=False,
        create_label_type=False,
        edit_labels=bool(env('LILAC_AUTH_USER_EDIT_LABELS', False)),
        label_all=not bool(env('LILAC_AUTH_USER_DISABLE_LABEL_ALL', False)),
        delete_rows=False,
        execute_remotely=False,
      ),
      concept=ConceptUserAccess(delete_any_concept=False),
    )

  return UserAccess(
    is_admin=is_admin,
    create_dataset=True,
    dataset=DatasetUserAccess(
      compute_signals=True,
      delete_dataset=True,
      delete_signals=True,
      update_settings=True,
      create_label_type=True,
      edit_labels=True,
      label_all=True,
      delete_rows=True,
      execute_remotely=has_garden_credentials(),
    ),
    concept=ConceptUserAccess(delete_any_concept=True),
  )



"""Utils for duckdb."""

import urllib.parse

import duckdb

from ..env import env


def duckdb_setup(con: duckdb.DuckDBPyConnection) -> None:
  """Setup DuckDB. This includes setting up performance optimizations."""
  con.execute(
    """
    SET enable_http_metadata_cache=true;
    SET enable_object_cache=true;
  """
  )


def convert_path_to_duckdb(filepath: str) -> str:
  """Convert a filepath to a duckdb filepath."""
  scheme = urllib.parse.urlparse(filepath).scheme
  options: dict[str, str] = {}
  if scheme == '':
    return filepath
  elif scheme == 'gs':
    options['s3_endpoint'] = 'storage.googleapis.com'
    if env('GCS_REGION'):
      options['s3_region'] = env('GCS_REGION')
    if env('GCS_ACCESS_KEY'):
      options['s3_access_key_id'] = env('GCS_ACCESS_KEY')
    if env('GCS_SECRET_KEY'):
      options['s3_secret_access_key'] = env('GCS_SECRET_KEY')
    filepath = filepath.replace('gs://', 's3://')
  elif scheme == 's3':
    if env('S3_ENDPOINT'):
      options['s3_endpoint'] = env('S3_ENDPOINT')
    if env('S3_REGION'):
      options['s3_region'] = env('S3_REGION')
    if env('S3_ACCESS_KEY'):
      options['s3_access_key_id'] = env('S3_ACCESS_KEY')
    if env('S3_SECRET_KEY'):
      options['s3_secret_access_key'] = env('S3_SECRET_KEY')
  else:
    raise ValueError(f'Unsupported scheme: {scheme}')
  if options:
    return f'{filepath}?{urllib.parse.urlencode(options, safe="+/")}'
  return filepath



"""Tests for the SQLite source."""
import os
import pathlib
import sqlite3

import pytest
from pytest_mock import MockerFixture

from ..schema import schema
from ..source import SourceSchema
from ..sources.sqlite_source import SQLiteSource


@pytest.fixture(autouse=True)
def set_project_path(tmp_path: pathlib.Path, mocker: MockerFixture) -> None:
  # We need to set the project path so extensions like https will be installed in temp dir.
  mocker.patch.dict(os.environ, {'LILAC_PROJECT_DIR': str(tmp_path)})


def test_sqlite(tmp_path: pathlib.Path) -> None:
  db_file = os.path.join(str(tmp_path), 'test.db')
  rows = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}, {'x': 3, 'y': 'thirty'}]
  conn = sqlite3.connect(db_file)
  cursor = conn.cursor()
  cursor.execute(
    """
    CREATE TABLE test (
        x INTEGER,
        y TEXT
    )
  """
  )
  # Insert sample data.
  for row in rows:
    cursor.execute('INSERT INTO test (x, y) VALUES (?, ?)', tuple(row.values()))
  conn.commit()
  conn.close()

  source = SQLiteSource(db_file=db_file, table='test')
  source.setup()
  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=3
  )

  items = list(source.yield_items())

  assert items == [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}, {'x': 3, 'y': 'thirty'}]



"""Tests for the JSON source."""
import json
import os
import pathlib

import pytest
from pytest_mock import MockerFixture

from ..schema import schema
from ..source import SourceSchema
from ..test_utils import retrieve_parquet_rows
from .json_source import JSONSource


@pytest.fixture(autouse=True)
def set_project_path(tmp_path: pathlib.Path, mocker: MockerFixture) -> None:
  # We need to set the project path so extensions like https will be installed in temp dir.
  mocker.patch.dict(os.environ, {'LILAC_PROJECT_DIR': str(tmp_path)})


def test_simple_json(tmp_path: pathlib.Path) -> None:
  json_records = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}]

  filename = 'test-dataset.jsonl'
  filepath = os.path.join(tmp_path, filename)
  with open(filepath, 'w') as f:
    f.write(json.dumps(json_records))

  source = JSONSource(filepaths=[filepath])
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=2
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}]


def test_simple_jsonl(tmp_path: pathlib.Path) -> None:
  json_records = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}]
  json_lines = [json.dumps(record) + '\n' for record in json_records]

  filename = 'test-dataset.jsonl'
  filepath = os.path.join(tmp_path, filename)
  with open(filepath, 'w') as f:
    f.writelines(json_lines)

  source = JSONSource(filepaths=[filepath])
  source.setup()

  source_schema = source.source_schema()

  assert source_schema == SourceSchema(
    fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=2
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)

  assert items == [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}]


def test_sampling_jsonl(tmp_path: pathlib.Path) -> None:
  json_records = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}, {'x': 3, 'y': 'thirty'}]
  json_lines = [json.dumps(record) + '\n' for record in json_records]

  filename = 'test-dataset.jsonl'
  filepath = os.path.join(tmp_path, filename)
  with open(filepath, 'w') as f:
    f.writelines(json_lines)

  source = JSONSource(filepaths=[filepath], sample_size=2)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=2
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert len(items) == 2


def test_sampling_greater_than_dataset(tmp_path: pathlib.Path) -> None:
  json_records = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}, {'x': 3, 'y': 'thirty'}]
  json_lines = [json.dumps(record) + '\n' for record in json_records]

  filename = 'test-dataset.jsonl'
  filepath = os.path.join(tmp_path, filename)
  with open(filepath, 'w') as f:
    f.writelines(json_lines)

  source = JSONSource(filepaths=[filepath], sample_size=4)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=3
  )

  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert len(items) == 3



"""Dataset formats are inferred from the dataset schema."""


from .default_formats import register_default_formats
from .openchat import OpenChat
from .sharegpt import ShareGPT

register_default_formats()

__all__ = ['ShareGPT', 'OpenChat']



"""Tests for dataset.select_groups()."""

import re
from datetime import datetime

import pytest
from pytest_mock import MockerFixture

from ..schema import Field, Item, MapType, field, schema
from . import dataset as dataset_module
from .dataset import GroupsSortBy, SelectGroupsResult, SortOrder
from .dataset_test_utils import TestDataMaker


def test_flat_data(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'name': 'Name1', 'age': 34, 'active': False},
    {'name': 'Name2', 'age': 45, 'active': True},
    {'age': 17, 'active': True},  # Missing "name".
    {'name': 'Name3', 'active': True},  # Missing "age".
    {'name': 'Name4', 'age': 55},  # Missing "active".
  ]
  dataset = make_test_data(items)

  result = dataset.select_groups(leaf_path='name')

  assert result.counts == [('Name1', 1), ('Name2', 1), ('Name3', 1), ('Name4', 1), (None, 1)]

  result = dataset.select_groups(leaf_path='age', bins=[20, 50, 60])
  assert result.counts == [('1', 2), ('0', 1), ('2', 1), (None, 1)]

  result = dataset.select_groups(leaf_path='active')
  assert result.counts == [
    (True, 3),
    (False, 1),
    (None, 1),  # Missing "active".
  ]


def test_flat_data_deleted(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'name': 'Name1', 'age': 34, 'active': False},
    {'name': 'Name2', 'age': 45, 'active': True},
    {'age': 17, 'active': True},  # Missing "name".
  ]
  dataset = make_test_data(items)

  result = dataset.select_groups(leaf_path='name')

  assert result.counts == [('Name1', 1), ('Name2', 1), (None, 1)]
  dataset.delete_rows(filters=[('name', 'equals', 'Name1')])

  result = dataset.select_groups(leaf_path='name')
  assert result.counts == [('Name2', 1), (None, 1)]


def test_result_counts(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'active': False},
    {'active': True},
    {'active': True},
    {'active': True},
    {},  # Missing "active".
  ]
  dataset = make_test_data(items, schema=schema({'active': 'boolean'}))

  result = dataset.select_groups(leaf_path='active')
  assert result.counts == [(True, 3), (False, 1), (None, 1)]


def test_order_by_value(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'active': False},
    {'active': False},
    {'active': True},
    {'active': False},
    {},  # Missing "active".
  ]
  dataset = make_test_data(items, schema=schema({'active': 'boolean'}))

  result = dataset.select_groups(leaf_path='active', sort_by=GroupsSortBy.VALUE)
  assert result.counts == [(True, 1), (False, 3), (None, 1)]

  result = dataset.select_groups(
    leaf_path='active', sort_by=GroupsSortBy.VALUE, sort_order=SortOrder.ASC
  )
  assert result.counts == [(False, 3), (True, 1), (None, 1)]


def test_list_of_structs(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'list_of_structs': [{'name': 'a'}, {'name': 'b'}]},
    {'list_of_structs': [{'name': 'c'}, {'name': 'a'}, {'name': 'd'}]},
    {'list_of_structs': [{'name': 'd'}]},
  ]
  dataset = make_test_data(items)

  result = dataset.select_groups(leaf_path='list_of_structs.*.name')
  assert result.counts == [('a', 2), ('d', 2), ('b', 1), ('c', 1)]


def test_nested_lists(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'nested_list': [[{'name': 'a'}], [{'name': 'b'}]]},
    {'nested_list': [[{'name': 'c'}, {'name': 'a'}], [{'name': 'd'}]]},
    {'nested_list': [[{'name': 'd'}]]},
  ]
  dataset = make_test_data(items)

  result = dataset.select_groups(leaf_path='nested_list.*.*.name')
  assert result.counts == [('a', 2), ('d', 2), ('b', 1), ('c', 1)]


def test_nested_struct(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'nested_struct': {'struct': {'name': 'c'}}},
    {'nested_struct': {'struct': {'name': 'b'}}},
    {'nested_struct': {'struct': {'name': 'a'}}},
  ]
  dataset = make_test_data(items)

  result = dataset.select_groups(leaf_path='nested_struct.struct.name')
  assert result.counts == [('a', 1), ('b', 1), ('c', 1)]


def test_named_bins(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'age': 34.0},
    {'age': 45.0},
    {'age': 17.0},
    {'age': 80.0},
    {'age': 55.0},
    {'age': float('nan')},
  ]
  dataset = make_test_data(items)

  result = dataset.select_groups(
    leaf_path='age',
    bins=[('young', None, 20), ('adult', 20, 50), ('middle-aged', 50, 65), ('senior', 65, None)],
  )
  assert result.counts == [('adult', 2), ('middle-aged', 1), ('senior', 1), ('young', 1), (None, 1)]


def test_schema_with_bins(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'age': 34},
    {'age': 45},
    {'age': 17},
    {'age': 80},
    {'age': 55},
    {'age': float('nan')},
  ]
  data_schema = schema(
    {
      'age': field(
        'float32',
        bins=[
          ('young', None, 20),
          ('adult', 20, 50),
          ('middle-aged', 50, 65),
          ('senior', 65, None),
        ],
      )
    }
  )
  dataset = make_test_data(items, data_schema)

  result = dataset.select_groups(leaf_path='age')
  assert result.counts == [('adult', 2), ('middle-aged', 1), ('senior', 1), ('young', 1), (None, 1)]


def test_filters(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'name': 'Name1', 'age': 34, 'active': False},
    {'name': 'Name2', 'age': 45, 'active': True},
    {'age': 17, 'active': True},  # Missing "name".
    {'name': 'Name3', 'active': True},  # Missing "age".
    {'name': 'Name4', 'age': 55},  # Missing "active".
  ]
  dataset = make_test_data(items)

  # active = True.
  result = dataset.select_groups(leaf_path='name', filters=[('active', 'equals', True)])
  assert result.counts == [('Name2', 1), ('Name3', 1), (None, 1)]

  # age < 35.
  result = dataset.select_groups(leaf_path='name', filters=[('age', 'less', 35)])
  assert result.counts == [('Name1', 1), (None, 1)]

  # age < 35 and active = True.
  result = dataset.select_groups(
    leaf_path='name', filters=[('age', 'less', 35), ('active', 'equals', True)]
  )
  assert result.counts == [(None, 1)]


def test_datetime(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'id': 1, 'date': datetime(2023, 1, 1)},
    {'id': 2, 'date': datetime(2023, 1, 15)},
    {'id': 3, 'date': datetime(2023, 2, 1)},
    {'id': 4, 'date': datetime(2023, 3, 1)},
    {
      'id': 5
      # Missing datetime.
    },
  ]
  dataset = make_test_data(items)
  result = dataset.select_groups('date')
  assert result.counts == [
    (datetime(2023, 1, 1), 1),
    (datetime(2023, 1, 15), 1),
    (datetime(2023, 2, 1), 1),
    (datetime(2023, 3, 1), 1),
    (None, 1),
  ]


def test_invalid_leaf(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'nested_struct': {'struct': {'name': 'c'}}},
    {'nested_struct': {'struct': {'name': 'b'}}},
    {'nested_struct': {'struct': {'name': 'a'}}},
  ]
  dataset = make_test_data(items)

  with pytest.raises(
    ValueError, match=re.escape('Leaf "(\'nested_struct\',)" not found in dataset')
  ):
    dataset.select_groups(leaf_path='nested_struct')

  with pytest.raises(
    ValueError, match=re.escape("Leaf \"('nested_struct', 'struct')\" not found in dataset")
  ):
    dataset.select_groups(leaf_path='nested_struct.struct')

  with pytest.raises(
    ValueError,
    match=re.escape("Path ('nested_struct', 'struct', 'wrong_name') not found in schema"),
  ):
    dataset.select_groups(leaf_path='nested_struct.struct.wrong_name')


def test_too_many_distinct(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  too_many_distinct = 5
  mocker.patch(f'{dataset_module.__name__}.TOO_MANY_DISTINCT', too_many_distinct)

  items: list[Item] = [{'feature': str(i)} for i in range(too_many_distinct + 10)]
  dataset = make_test_data(items)

  res = dataset.select_groups('feature')
  assert res.too_many_distinct is True
  assert res.counts == []


def test_auto_bins_for_float(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [{'feature': float(i)} for i in range(5)] + [{'feature': float('nan')}]
  dataset = make_test_data(items)

  res = dataset.select_groups('feature')
  assert res.counts == [('0', 1), ('11', 1), ('14', 1), ('3', 1), ('7', 1), (None, 1)]
  assert res.too_many_distinct is False
  assert res.bins


def test_map_dtype(make_test_data: TestDataMaker) -> None:
  items = [
    {'column': {'a': 1.0, 'b': 2.0}},
    {'column': {'b': 2.5}},
    {'column': {'a': 3.0, 'c': 3.5}},
  ]
  data_schema = schema(
    {'column': Field(dtype=MapType(key_type='string', value_field=field('float32')))}
  )
  dataset = make_test_data(items, schema=data_schema)
  assert dataset.select_groups('column.a', bins=[0, 2]) == SelectGroupsResult(
    too_many_distinct=False,
    counts=[('1', 1), ('2', 1)],
    bins=[('0', None, 0), ('1', 0, 2), ('2', 2, None)],
  )
  assert dataset.select_groups('column.b', bins=[0, 1.5]) == SelectGroupsResult(
    too_many_distinct=False,
    counts=[('2', 2)],
    bins=[('0', None, 0), ('1', 0, 1.5), ('2', 1.5, None)],
  )
  assert dataset.select_groups('column.c', bins=[0, 2]) == SelectGroupsResult(
    too_many_distinct=False,
    counts=[('2', 1)],
    bins=[('0', None, 0), ('1', 0, 2), ('2', 2, None)],
  )
  with pytest.raises(ValueError, match='Cannot compute groups on a map field'):
    dataset.select_groups('column')



"""Tests for `db.select_rows_schema()`."""

from typing import ClassVar, Iterable, Iterator, Optional, cast

import numpy as np
import pytest
from typing_extensions import override

from ..schema import (
  EMBEDDING_KEY,
  PATH_WILDCARD,
  Field,
  Item,
  RichData,
  SignalInputType,
  SpanVector,
  field,
  lilac_embedding,
  schema,
  span,
)
from ..signal import (
  TextEmbeddingSignal,
  TextSignal,
  VectorSignal,
  clear_signal_registry,
  register_signal,
)
from ..signals.concept_labels import ConceptLabelsSignal
from ..signals.concept_scorer import ConceptSignal
from ..signals.semantic_similarity import SemanticSimilaritySignal
from ..signals.substring_search import SubstringSignal
from .dataset import (
  Column,
  ConceptSearch,
  KeywordSearch,
  SearchResultInfo,
  SelectRowsSchemaResult,
  SelectRowsSchemaUDF,
  SemanticSearch,
  SortOrder,
  SortResult,
)
from .dataset_test_utils import TestDataMaker

TEST_DATA: list[Item] = [
  {
    'erased': False,
    'people': [
      {
        'name': 'A',
        'zipcode': 0,
        'locations': [{'city': 'city1', 'state': 'state1'}, {'city': 'city2', 'state': 'state2'}],
      }
    ],
  },
  {
    'erased': True,
    'people': [
      {
        'name': 'B',
        'zipcode': 1,
        'locations': [{'city': 'city3', 'state': 'state3'}, {'city': 'city4'}, {'city': 'city5'}],
      },
      {'name': 'C', 'zipcode': 2, 'locations': [{'city': 'city1', 'state': 'state1'}]},
    ],
  },
]


class TestSplitter(TextSignal):
  """Split documents into sentence by splitting on period."""

  name: ClassVar[str] = 'test_splitter'

  @override
  def fields(self) -> Field:
    return field(fields=['string_span'])

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    for text in data:
      if not isinstance(text, str):
        raise ValueError(f'Expected text to be a string, got {type(text)} instead.')
      sentences = [f'{sentence.strip()}.' for sentence in text.split('.') if sentence]
      yield [
        span(text.index(sentence), text.index(sentence) + len(sentence)) for sentence in sentences
      ]


EMBEDDINGS: list[tuple[str, list[float]]] = [
  ('hello.', [1.0, 0.0, 0.0]),
  ('hello2.', [1.0, 1.0, 0.0]),
  ('hello world.', [1.0, 1.0, 1.0]),
  ('hello world2.', [2.0, 1.0, 1.0]),
]

STR_EMBEDDINGS: dict[str, list[float]] = {text: embedding for text, embedding in EMBEDDINGS}


class TestEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      yield [lilac_embedding(0, len(example), np.array(STR_EMBEDDINGS[cast(str, example)]))]


class TestEmbeddingSumSignal(VectorSignal):
  """Sums the embeddings to return a single floating point value."""

  name: ClassVar[str] = 'test_embedding_sum'
  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  @override
  def fields(self) -> Field:
    return field('float32')

  @override
  def vector_compute(
    self, all_vector_spans: Iterable[list[SpanVector]]
  ) -> Iterator[Optional[Item]]:
    # The signal just sums the values of the embedding.
    for vector_spans in all_vector_spans:
      yield vector_spans[0]['vector'].sum()


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_signal(LengthSignal)
  register_signal(AddSpaceSignal)
  register_signal(TestSplitter)
  register_signal(TestEmbedding)
  register_signal(TestEmbeddingSumSignal)

  # Unit test runs.
  yield

  # Teardown.
  clear_signal_registry()


class LengthSignal(TextSignal):
  name: ClassVar[str] = 'length_signal'

  def fields(self) -> Field:
    return field('int32')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield len(text_content)


class AddSpaceSignal(TextSignal):
  name: ClassVar[str] = 'add_space_signal'

  def fields(self) -> Field:
    return field('string')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield cast(str, text_content) + ' '


def test_simple_schema(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)
  result = dataset.select_rows_schema(combine_columns=True)
  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'erased': 'boolean',
        'people': [
          {
            'name': 'string',
            'zipcode': 'int32',
            'locations': [{'city': 'string', 'state': 'string'}],
          }
        ],
      }
    )
  )


def test_subselection_with_combine_cols(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  result = dataset.select_rows_schema(
    [('people', '*', 'zipcode'), ('people', '*', 'locations', '*', 'city')], combine_columns=True
  )
  assert result == SelectRowsSchemaResult(
    data_schema=schema({'people': [{'zipcode': 'int32', 'locations': [{'city': 'string'}]}]})
  )

  result = dataset.select_rows_schema(
    [('people', '*', 'name'), ('people', '*', 'locations')], combine_columns=True
  )
  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {'people': [{'name': 'string', 'locations': [{'city': 'string', 'state': 'string'}]}]}
    )
  )

  result = dataset.select_rows_schema([('people', '*')], combine_columns=True)
  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'people': [
          {
            'name': 'string',
            'zipcode': 'int32',
            'locations': [{'city': 'string', 'state': 'string'}],
          }
        ]
      }
    )
  )


def test_udf_with_combine_cols(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  length_signal = LengthSignal()
  result = dataset.select_rows_schema(
    [
      ('people', '*', 'locations', '*', 'city'),
      Column(('people', '*', 'name'), signal_udf=length_signal),
    ],
    combine_columns=True,
  )
  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'people': [
          {
            'name': {'length_signal': field('int32', length_signal.model_dump(exclude_none=True))},
            'locations': [{'city': 'string'}],
          }
        ]
      }
    ),
    udfs=[SelectRowsSchemaUDF(path=('people', '*', 'name', length_signal.key()))],
  )


def test_embedding_udf_with_combine_cols(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  add_space_signal = AddSpaceSignal()
  path = ('people', '*', 'name')
  dataset.compute_signal(add_space_signal, path)
  result = dataset.select_rows_schema(
    [path, Column(path, signal_udf=add_space_signal)], combine_columns=True
  )
  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'people': [
          {
            'name': field(
              'string',
              fields={
                'add_space_signal': field(
                  'string', signal=add_space_signal.model_dump(exclude_none=True)
                )
              },
            )
          }
        ]
      }
    ),
    udfs=[SelectRowsSchemaUDF(path=(*path, add_space_signal.key()))],
  )


def test_udf_chained_with_combine_cols(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello. hello2.'}, {'text': 'hello world. hello world2.'}])

  test_splitter = TestSplitter()
  dataset.compute_signal(test_splitter, ('text'))
  add_space_signal = AddSpaceSignal()
  result = dataset.select_rows_schema(
    [('text'), Column(('text'), signal_udf=add_space_signal)], combine_columns=True
  )

  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'add_space_signal': field('string', add_space_signal.model_dump(exclude_none=True)),
            'test_splitter': field(
              signal=test_splitter.model_dump(exclude_none=True), fields=['string_span']
            ),
          },
        )
      }
    ),
    udfs=[SelectRowsSchemaUDF(path=('text', add_space_signal.key()))],
  )


def test_search_keyword_schema(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world', 'text2': 'hello world2'}])
  query_world = 'world'
  query_hello = 'hello'

  result = dataset.select_rows_schema(
    searches=[
      KeywordSearch(path='text', query=query_world),
      KeywordSearch(path='text2', query=query_hello),
    ],
    combine_columns=True,
  )

  expected_world_signal = SubstringSignal(query=query_world)
  expected_hello_signal = SubstringSignal(query=query_hello)

  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            expected_world_signal.key(): field(
              signal=expected_world_signal.model_dump(exclude_none=True), fields=['string_span']
            )
          },
        ),
        'text2': field(
          'string',
          fields={
            expected_hello_signal.key(): field(
              signal=expected_hello_signal.model_dump(exclude_none=True), fields=['string_span']
            )
          },
        ),
      }
    ),
    search_results=[
      SearchResultInfo(
        search_path=('text',), result_path=('text', expected_world_signal.key(), PATH_WILDCARD)
      ),
      SearchResultInfo(
        search_path=('text2',), result_path=('text2', expected_hello_signal.key(), PATH_WILDCARD)
      ),
    ],
    udfs=[
      SelectRowsSchemaUDF(path=('text', expected_world_signal.key())),
      SelectRowsSchemaUDF(path=('text2', expected_hello_signal.key())),
    ],
  )


def test_search_semantic_schema(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world.'}])
  query_world = 'world'

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  result = dataset.select_rows_schema(
    searches=[SemanticSearch(path='text', query=query_world, embedding='test_embedding')],
    combine_columns=True,
  )

  test_embedding = TestEmbedding()
  expected_world_signal = SemanticSimilaritySignal(query=query_world, embedding='test_embedding')

  similarity_score_path = ('text', expected_world_signal.key())
  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'test_embedding': field(
              signal=test_embedding.model_dump(exclude_none=True),
              fields=[field('string_span', fields={EMBEDDING_KEY: 'embedding'})],
            ),
            expected_world_signal.key(): field(
              signal=expected_world_signal.model_dump(exclude_none=True),
              fields=[field('string_span', fields={'score': 'float32'})],
            ),
          },
        )
      }
    ),
    udfs=[SelectRowsSchemaUDF(path=similarity_score_path)],
    search_results=[SearchResultInfo(search_path=('text',), result_path=similarity_score_path)],
    sorts=[
      SortResult(
        path=(*similarity_score_path, PATH_WILDCARD, 'score'), order=SortOrder.DESC, search_index=0
      )
    ],
  )


def test_search_concept_schema(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world.'}])

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  result = dataset.select_rows_schema(
    searches=[
      ConceptSearch(
        path='text',
        concept_namespace='test_namespace',
        concept_name='test_concept',
        embedding='test_embedding',
      )
    ],
    combine_columns=True,
  )

  expected_world_signal = ConceptSignal(
    namespace='test_namespace', concept_name='test_concept', embedding='test_embedding'
  )
  expected_labels_signal = ConceptLabelsSignal(
    namespace='test_namespace', concept_name='test_concept'
  )

  concept_score_path = ('text', expected_world_signal.key())
  concept_labels_path = ('text', expected_labels_signal.key())
  assert result == SelectRowsSchemaResult(
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'test_embedding': field(
              signal=test_embedding.model_dump(exclude_none=True),
              fields=[field('string_span', fields={EMBEDDING_KEY: 'embedding'})],
            ),
            expected_world_signal.key(): field(
              signal=expected_world_signal.model_dump(exclude_none=True),
              fields=[
                field(
                  dtype='string_span',
                  fields={
                    'score': field(
                      'float32', bins=[('Not in concept', None, 0.5), ('In concept', 0.5, None)]
                    )
                  },
                )
              ],
            ),
            'test_namespace/test_concept/labels/preview': field(
              fields=[field('string_span', fields={'label': 'boolean', 'draft': 'string'})],
              signal=expected_labels_signal.model_dump(exclude_none=True),
            ),
          },
        )
      }
    ),
    udfs=[
      SelectRowsSchemaUDF(path=concept_labels_path),
      SelectRowsSchemaUDF(path=concept_score_path),
    ],
    search_results=[
      SearchResultInfo(search_path=('text',), result_path=concept_labels_path),
      SearchResultInfo(search_path=('text',), result_path=concept_score_path),
    ],
    sorts=[
      SortResult(
        path=(*concept_score_path, PATH_WILDCARD, 'score'), order=SortOrder.DESC, search_index=0
      )
    ],
  )


def test_search_sort_override(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello world.'}])
  query_world = 'world'

  test_embedding = TestEmbedding()
  dataset.compute_signal(test_embedding, ('text'))

  result = dataset.select_rows_schema(
    searches=[SemanticSearch(path='text', query=query_world, embedding='test_embedding')],
    # Explicit sort by overrides the semantic search.
    sort_by=[('text',)],
    sort_order=SortOrder.DESC,
    combine_columns=True,
  )

  assert result.sorts == [SortResult(path=('text',), order=SortOrder.DESC)]
